"""Abstract Base Class for all Sessions."""
from __future__ import annotations

import itertools
import logging
from datetime import datetime
from functools import wraps
from inspect import isclass
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from rdflib import OWL, RDF, RDFS, SKOS, BNode, Graph, Literal, URIRef
from rdflib.graph import ModificationException, ReadOnlyGraphAggregate
from rdflib.plugins.sparql.processor import SPARQLResult
from rdflib.query import ResultRow
from rdflib.term import Identifier, Node, Variable

from simphony_osp.ontology.annotation import OntologyAnnotation
from simphony_osp.ontology.attribute import OntologyAttribute
from simphony_osp.ontology.entity import OntologyEntity
from simphony_osp.ontology.individual import (
    MultipleResultsError,
    OntologyIndividual,
    ResultEmptyError,
)
from simphony_osp.ontology.namespace import OntologyNamespace
from simphony_osp.ontology.oclass import OntologyClass
from simphony_osp.ontology.parser import OntologyParser
from simphony_osp.ontology.relationship import OntologyRelationship
from simphony_osp.ontology.utils import DataStructureSet, compatible_classes
from simphony_osp.utils import simphony_namespace
from simphony_osp.utils.cache import lru_cache_weak
from simphony_osp.utils.datatypes import (
    UID,
    AnnotationValue,
    AttributeValue,
    RelationshipValue,
    Triple,
)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from simphony_osp.interfaces.interface import Interface, InterfaceDriver

ENTITY = TypeVar("ENTITY", bound=OntologyEntity)


RDF_type = RDF.type
OWL_inverseOf = OWL.inverseOf


class Environment:
    """Environment where ontology entities may be created.

    E.g. sessions, containers.
    """

    # ↓ --------------------- Public API --------------------- ↓ #

    @property
    def locked(self) -> bool:
        """Whether the environment is locked or not.

        A locked environment will not be closed when using it as a context
        manager and leaving the context. Useful for setting it as the
        default environment when it is not intended to close it afterwards.
        """
        return (self._lock + bool(self._user_lock)) > 0

    @locked.setter
    def locked(self, value: bool):
        """Lock or unlock an environment."""
        self._user_lock = value

    def __enter__(self):
        """Set this as the default environment."""
        self._stack_default_environment.append(self)
        self._environment_references.add(self)
        return self

    def __exit__(self, *args):
        """Set the default environment back to the previous default."""
        if self is not self._stack_default_environment[-1]:
            raise RuntimeError(
                "Trying to exit the an environment context "
                "manager which was not the last entered one."
            )
        self._stack_default_environment.pop()
        if (
            self not in self._stack_default_environment
            and not self.subscribers
            and not self.locked
        ):
            self.close()
        return False

    def close(self):
        """Close this environment."""
        for environment in self.subscribers:
            environment.close()
            self.subscribers.remove(environment)
        self._environment_references.remove(self)

    def __bool__(self) -> bool:
        """Evaluate the truth value of the environment.

        Such value is always true.
        """
        return True

    # ↑ --------------------- Public API --------------------- ↑ #

    _session_linked: Optional[Session] = None
    _stack_default_environment: List[Environment] = []
    _environment_references: Set[Environment] = set()

    _lock: int = 0
    """See the docstring of `locked` for an explanation of what locking an
    environment means."""

    _user_lock: bool = False
    """See the docstring of `locked` for an explanation of what locking an
    environment means.

    This property manages locks performed by the user (by setting the
    `locked` attribute).
    """

    def lock(self):
        """Increase the lock count.

        This way of locking is not meant to be used by end users, only
        internally within SimPhoNy code, as it allows to lock the
        environment several times, later requiring several unlocks, which is
        unintuitive.

        See the docstring of `locked` for an explanation of what locking an
        environment means.
        """
        self._lock += 1

    def unlock(self):
        """Decrease the lock count.

        This way of unlocking is not meant to be used by end users, only
        internally within SimPhoNy code, as it is the counterpart of
        `lock`, which allows to lock the environment several times,
        later requiring several unlocks, which is unintuitive.

        See the docstring of `locked` for an explanation of what locking an
        environment means.
        """
        self._lock = self._lock - 1 if self._lock > 0 else 0

    _subscribers: Set[Environment]
    """A private attribute is used in order not to interfere with the
    `__getattr__`method from OntologyIndividual."""

    def __init__(self, *args, **kwargs):
        """Initialize the environment with an empty set of subscribers."""
        self._subscribers = set()
        super().__init__(*args, **kwargs)

    @property
    def subscribers(self) -> Set[Environment]:
        """Environments that depend on this instance.

        Such environments will be closed when this instance is closed.
        """
        return self._subscribers

    @subscribers.setter
    def subscribers(self, value: Set[Environment]):
        """Setter for the private  `_subscribers` attribute."""
        self._subscribers = value

    @classmethod
    def get_default_environment(cls) -> Optional[Environment]:
        """Returns the default environment."""
        for environment in cls._stack_default_environment[::-1]:
            return environment
        else:
            return None


class SessionSet(DataStructureSet):
    """A set interface to a session.

    This class looks like and acts like the standard `set`, but it is an
    interface to the methods from `Session` that manage the addition and
    removal of individuals.

    This class does not hold any information itself, thus it is safe to
    spawn multiple instances linked to the same session (when
    single-threading).
    """

    _session: Session

    def __init__(
        self,
        session: Optional[Session] = None,
        oclass: Optional[OntologyClass] = None,
        uids: Optional[Iterable[UID]] = None,
    ):
        """Fix the linked session, the class and the identifier filter."""
        if oclass is not None and not isinstance(oclass, OntologyClass):
            raise TypeError(
                "Found object of type %s passed to argument "
                "oclass. Should be an OntologyClass." % type(oclass)
            )
        uids = tuple(uids) if uids is not None else None
        if uids is not None:
            for uid in uids:
                if not isinstance(uid, UID):
                    raise TypeError(
                        "Found object of type %s. Should be an UID."
                        % type(uid)
                    )

        self._class_filter = oclass
        self._uid_filter = uids
        self._session = session or Session.get_default_session()
        super().__init__()

    def __iter__(self):
        """The entities contained in the session."""
        identifiers = self._uid_filter
        class_ = self._class_filter

        if identifiers:
            yielded = set()
            for entity in self._iter_identifiers():
                if entity not in yielded:
                    yield entity
        elif class_:
            yield from (
                row[0]
                for row in self._session.sparql(
                    f"""
                    SELECT DISTINCT ?entity WHERE {{
                        ?entity rdf:type/rdfs:subClassOf*
                        <{class_.iri}> .
                    }}
                """,
                    ontology=True,
                )(entity=OntologyIndividual)
            )
        else:
            yield from iter(self._session)

    def __contains__(self, item: OntologyIndividual) -> bool:
        """Check whether an ontology entity belongs to the session."""
        return item in self._session and (
            item.is_a(self._class_filter) if self._class_filter else True
        )

    def update(self, other: Iterable[OntologyIndividual]) -> None:
        """Update the set with the union of itself and others."""
        other = set(other)
        if self._class_filter:
            for individual in other:
                if not individual.is_a(self._class_filter):
                    raise RuntimeError(
                        f"Cannot update {self} with {individual} because it "
                        f"does not belong to class {self._class_filter}."
                    )

        self._session.add(
            *other,
            merge=True,
            exists_ok=True,
        )

    def intersection_update(self, other: Iterable[OntologyIndividual]) -> None:
        """Update the set with the intersection of itself and another."""
        intersection = set(
            x
            for x in other
            if (x.identifier, RDF.type, None) in self._session.graph
        )

        if self._class_filter:
            for individual in intersection:
                if not individual.is_a(self._class_filter):
                    raise RuntimeError(
                        f"Cannot update {self} with {individual} because it "
                        f"does not belong to class {self._class_filter}."
                    )

        existing = set(
            x.identifier for x in self._session.get(oclass=self._class_filter)
        )
        remove = existing - set(x.identifier for x in intersection)

        self._session.add(
            *intersection,
            merge=True,
            exists_ok=True,
        )
        self._session.delete(remove)

    def difference_update(self, other: Iterable[OntologyIndividual]) -> None:
        """Remove all elements of another set from this set."""
        other = set(other)
        exists = set()
        for entity in other:
            try:
                exists.add(
                    self._session.from_identifier_typed(
                        entity.identifier, typing=OntologyIndividual
                    )
                )
            except KeyError:
                pass

        if self._class_filter:
            for individual in exists:
                if not individual.is_a(self._class_filter):
                    raise RuntimeError(
                        f"Cannot delete {individual} because it "
                        f"does not belong to class {self._class_filter}."
                    )

        self._session.delete(exists)

    def symmetric_difference_update(
        self, other: Iterable[OntologyIndividual]
    ) -> None:
        """Update set with the symmetric difference of it and another."""
        other = set(other)
        intersection = set(
            x
            for x in other
            if (x.identifier, RDF.type, None) in self._session.graph
        )
        add = other - intersection
        delete = set(
            x
            for x in self
            if x.identifier in (x.identifier for x in intersection)
        )

        if self._class_filter:
            for individual in add:
                if not individual.is_a(self._class_filter):
                    raise RuntimeError(
                        f"Cannot add {individual} because it "
                        f"does not belong to class {self._class_filter}."
                    )
            for individual in delete:
                if not individual.is_a(self._class_filter):
                    raise RuntimeError(
                        f"Cannot delete {individual} because it "
                        f"does not belong to class {self._class_filter}."
                    )

        self._session.add(add, merge=False, exists_ok=False)
        self._session.delete(delete)

    def __repr__(self) -> str:
        """Return repr(self)."""
        return (
            set(self).__repr__()
            + " <"
            + (
                f"class {self._class_filter} "
                if self._class_filter is not None
                else ""
            )
            + f"of session {self._session.identifier or self._session}>"
        )

    def one(
        self,
    ) -> OntologyIndividual:
        """Return one element.

        Return one element if the set contains one element, else raise
        an exception.

        Returns:
            The only element contained in the set.

        Raises:
            ResultEmptyError: No elements in the set.
            MultipleResultsError: More than one element in the set.
        """
        iter_self = iter(self)
        first_element = next(iter_self, StopIteration)
        if first_element is StopIteration:
            raise ResultEmptyError("No elements to be yielded.")
        second_element = next(iter_self, StopIteration)
        if second_element is not StopIteration:
            raise MultipleResultsError("More than one element can be yielded.")
        return first_element

    def any(
        self,
    ) -> Optional[Union[AnnotationValue, AttributeValue, RelationshipValue]]:
        """Return any element of the set.

        Returns:
            Any element from the set if the set is not empty, else None.
        """
        return next(iter(self), None)

    def all(self) -> SessionSet:
        """Return all elements from the set.

        Returns:
            All elements from the set, namely the set itself.
        """
        return self

    def _iter_identifiers(self) -> Iterator[Optional[OntologyIndividual]]:
        identifiers = self._uid_filter
        class_ = self._class_filter
        for i, identifier in identifiers:
            try:
                entity = self._session.from_identifier_typed(
                    identifier, OntologyIndividual
                )
            except KeyError:
                entity = None
            if entity and class_ and not entity.is_a(class_):
                entity = None
            yield entity


class Session(Environment):
    """Interface to a Graph containing OWL ontology entities."""

    # ↓ --------------------- Public API --------------------- ↓ #
    """These methods are meant to be available to the end-user."""

    identifier: Optional[str] = None
    """A label for the session.

    The identifier is just a label for the session to be displayed within
    Python (string representation of the session). It has no other effect.
    """

    @property
    def ontology(self) -> Session:
        """Another session considered to be the T-Box of this one.

        In a normal setting, a session is considered only to contain an A-Box.
        When it is necessary to look for a class, a relationship, an attribute
        or an annotation property, the session will look there for their
        definition.
        """
        return self._ontology or Session.default_ontology

    @ontology.setter
    def ontology(self, value: Optional[Session]) -> None:
        """Set the T-Box of this session."""
        if not isinstance(value, (Session, type(None))):
            raise TypeError(
                f"Expected {Session} or {type(None)}, not type {value}."
            )
        self._ontology = value

    label_properties: Tuple[URIRef] = (SKOS.prefLabel, RDFS.label)
    """The identifiers of the RDF predicates to be considered as labels.

    The entity labels are used, for example, to be able to get ontology
    entities from namespace or session objects by such label.

    The order in which the properties are specified in the tuple matters. To
    determine the label of an object, the properties will be checked from
    left to right, until one of them is defined for that specific entity.
    This will be the label of such ontology entity. The rest of the
    properties to the right of such property will be ignored for that
    entity.

    For example, in the default case above, if an entity has an
    `SKOS.prefLabel` it will be considered to be its label, even if it also
    has an `RDFS.label`, which will be ignored. If another entity has no
    `SKOS.prefLabel` but has a `RDFS.label`, then the `RDFS.label` will
    define its label. This means that for some entity, one label property
    may be used while for another, a different property can be in use. If
    none of the properties are defined, then the entity is considered to
    have no label.
    """

    label_languages: Tuple[URIRef] = ("en",)
    # TODO: Set to user's language preference from the OS (users can usually
    #  set such a list in modern operating systems).
    """The preferred languages for the default label.

    Normally, entities will be available from all languages. However,
    in some places the label has to be printed. In such cases this default
    label will be used.

    When defining the label for an object as described in the
    `label_properties` docstring above, this list will also be checked from
    left to right. When one of the languages specified is available,
    this will define the default label. Then the default label will default to
    english. If also not available, then any language will be used.
    """

    def commit(self) -> None:
        """Commit pending changes to the session's graph."""
        self._graph.commit()
        # if self.ontology is not self:
        #    self.ontology.commit()
        self.creation_set = set()

    def compute(self, **kwargs) -> None:
        """Run simulations on supported graph stores."""
        from simphony_osp.interfaces.remote.client import RemoteStoreClient

        if self._driver is not None:
            self.commit()
            self._driver.compute(**kwargs)
        elif isinstance(self._graph.store, RemoteStoreClient):
            self._graph.store.execute_method("compute")
        else:
            raise AttributeError(
                f"Session {self} is not attached to a "
                f"simulation engine. Thus, the attribute "
                f"`compute` is not available."
            )

    def close(self) -> None:
        """Close the connection to the session's backend.

        Sessions are an interface to a graph linked to an RDFLib store (a
        backend). If the session will not be used anymore, then it makes
        sense to close the connection to such backend to free resources.
        """
        if self in self._stack_default_environment:
            raise RuntimeError(
                "Cannot close a session that is currently "
                "being used as a context manager."
            )
        super().close()
        self.graph.close(commit_pending_transaction=False)

    def sparql(self, query: str, ontology: bool = False) -> QueryResult:
        """Perform a SPARQL CONSTRUCT, DESCRIBE, SELECT or ASK query.

        The query is performed on the session's data (the ontology is not
        included).

        Args:
            query: String to use as query.
            ontology: Whether to include the ontology in the query or not.
                When the ontology is included, only read-only queries are
                possible.
        """
        graph = (
            self.graph
            if not ontology
            else ReadOnlyGraphAggregate([self.graph, self.ontology.graph])
        )
        result = graph.query(query)
        return QueryResult(
            {
                "type_": result.type,
                "vars_": result.vars,
                "bindings": result.bindings,
                "askAnswer": result.askAnswer,
                "graph": result.graph,
            },
            session=self,
        )

    def __enter__(self):
        """Sets the session as the default session."""
        super().__enter__()
        self.creation_set = set()
        return self

    def __contains__(self, item: OntologyEntity):
        """Check whether an ontology entity is stored on the session."""
        return item.session is self

    def __iter__(self) -> Iterator[OntologyEntity]:
        """Iterate over all the ontology entities in the session.

        This operation can be computationally VERY expensive.
        """
        # Warning: entities can be repeated.
        return (
            self.from_identifier(identifier)
            for identifier in self.iter_identifiers()
        )

    def __len__(self) -> int:
        """Return the number of ontology entities within the session."""
        return sum(1 for _ in self)

    def __str__(self):
        """Convert the session to a string."""
        # TODO: Return the kind of RDFLib store attached too.
        return (
            f"<{self.__class__.__module__}.{self.__class__.__name__}: "
            f"{self.identifier if self.identifier is not None else ''} "
            f"at {hex(id(self))}>"
        )

    @lru_cache_weak(maxsize=4096)
    # On `__init__.py` there is an option to bypass this cache when the
    # session is not a T-Box.
    def from_identifier(self, identifier: Node) -> OntologyEntity:
        """Get an ontology entity from its identifier.

        Args:
            identifier: The identifier of the entity.

        Raises:
            KeyError: The ontology entity is not stored in this session.

        Returns:
            The OntologyEntity.
        """
        # WARNING: This method is a central point in SimPhoNy. Change with
        #  care.
        # TIP: Since the method is a central point in SimPhoNy, any
        #  optimization it gets will speed up SimPhoNy, while bad code in
        #  this method will slow it down.

        # Look for embedded classes.
        compatible = {
            rdf_type: compatible_classes(rdf_type, identifier)
            for rdf_type in self._graph.objects(identifier, RDF_type)
        }

        # If not an embedded class, then the type may be known in
        # the ontology. This means that an ontology individual would
        # have to be spawned.
        for rdf_type, found in compatible.items():
            if not found:
                try:
                    self.ontology.from_identifier(rdf_type)
                    found |= {OntologyIndividual}
                    break
                except KeyError:
                    pass

        compatible = set().union(*compatible.values())

        if (
            OntologyRelationship not in compatible
            and (identifier, OWL_inverseOf, None) in self._graph
        ):
            compatible |= {OntologyRelationship}

        """Some ontologies are hybrid RDFS and OWL ontologies (i.e. FOAF).
        In such cases, object and datatype properties are preferred to
        annotation properties."""
        if OntologyAnnotation in compatible and (
            compatible & {OntologyRelationship, OntologyAttribute}
        ):
            compatible.remove(OntologyAnnotation)

        """Finally return the single compatible class or raise an exception."""
        if len(compatible) >= 2:
            raise RuntimeError(
                f"Two or more python classes ("
                f"{', '.join(map(str, compatible))}) "
                f"could be spawned from {identifier}."
            )
        try:
            python_class = compatible.pop()
            return python_class(uid=UID(identifier), session=self, merge=None)
        except KeyError:
            raise KeyError(
                f"Identifier {identifier} does not match any OWL "
                f"entity, any entity natively supported by "
                f"SimPhoNy, nor an ontology individual "
                f"belonging to a class in the ontology."
            )

    def from_identifier_typed(
        self, identifier: Node, typing: Type[ENTITY]
    ) -> ENTITY:
        """Get an ontology entity from its identifier, enforcing a type check.

        Args:
            identifier: The identifier of the entity.
            typing: The expected type of the ontology entity matching the
                given identifier.

        Raises:
            KeyError: The ontology entity is not stored in this session.

        Returns:
            The OntologyEntity.
        """
        entity = self.from_identifier(identifier)
        if not isinstance(entity, typing):
            raise TypeError(f"{identifier} is not of class {typing}.")
        return entity

    @lru_cache_weak(maxsize=4096)
    # On `__init__.py` there is an option to bypass this cache when the
    # session is not a T-Box.
    def from_label(
        self,
        label: str,
        lang: Optional[str] = None,
        case_sensitive: bool = False,
    ) -> FrozenSet[OntologyEntity]:
        """Get an ontology entity from its label.

        Args:
            label: The label of the ontology entity.
            lang: The language of the label.
            case_sensitive: when false, look for similar labels with
                different capitalization.

        Raises:
            KeyError: Unknown label.

        Returns:
            The ontology entity.
        """
        results = set()

        identifiers_and_labels = self.iter_labels(
            lang=lang,
            return_prop=False,
            return_literal=False,
            return_identifier=True,
        )
        if case_sensitive is False:
            comp_label = label.lower()
            identifiers_and_labels = (
                (label.lower(), identifier)
                for label, identifier in identifiers_and_labels
            )
        else:
            comp_label = label

        identifiers_and_labels = (
            (label, identifier)
            for label, identifier in identifiers_and_labels
            if label == comp_label
        )

        for _, identifier in identifiers_and_labels:
            try:
                results.add(self.from_identifier(identifier))
            except KeyError:
                pass
        if len(results) == 0:
            error = "No element with label %s was found in ontology %s." % (
                label,
                self,
            )
            raise KeyError(error)

        return frozenset(results)

    def add(
        self,
        *individuals: Union[OntologyIndividual, Iterable[OntologyIndividual]],
        merge: bool = False,
        exists_ok: bool = False,
    ) -> Union[OntologyIndividual, FrozenSet[OntologyIndividual]]:
        """Copies the ontology entities to the session."""
        # Unpack iterables
        individuals = list(
            individual
            for x in individuals
            for individual in (
                x if not isinstance(x, OntologyIndividual) else (x,)
            )
        )
        # Get the identifiers of the individuals
        identifiers = list(individual.identifier for individual in individuals)
        # Get a list of files within the individuals to add
        files = {
            individual
            for individual in individuals
            if set(class_.identifier for class_ in individual.superclasses)
            & {simphony_namespace.File}
            and individual.session is not self
        }

        # Paste the individuals
        """The attributes of the individuals are always kept. The
        relationships between the individuals are only kept when they are
        pasted together.
        """
        if (
            any(
                (identifier, None, None) in self.graph
                for identifier in identifiers
            )
            and exists_ok is False
        ):
            raise RuntimeError(
                "Some of the added entities already exist on the session."
            )
        elif (
            merge
            and files
            and any(
                (identifier, None, None) in self.graph
                for identifier in {x.identifier for x in files}
            )
        ):
            raise RuntimeError(
                "Some of the added file entities already exist on the "
                "session. File entities cannot be merged with existing ones."
            )
        delete = (
            (individual.identifier, None, None)
            for individual in individuals
            if individual.session is not self
        )
        add = (
            (s, p, o)
            for individual in individuals
            for s, p, o in individual.session.graph.triples(
                (individual.identifier, None, None)
            )
            if (p == RDF.type or isinstance(o, Literal) or o in identifiers)
        )
        if not merge:
            """Replace previous individuals if merge is False."""
            for pattern in delete:
                self.graph.remove(pattern)
        self.graph.addN((s, p, o, self.graph) for s, p, o in add)
        files = ((file.identifier, file.operations.handle) for file in files)
        for identifier, contents in files:
            self.from_identifier_typed(
                identifier, typing=OntologyIndividual
            ).operations.overwrite(contents)
        added_objects = list(
            self.from_identifier_typed(identifier, typing=OntologyIndividual)
            for identifier in identifiers
        )
        return (
            next(iter(added_objects), None)
            if len(added_objects) <= 1
            else added_objects
        )

    def delete(
        self,
        *entities: Union[
            Union[OntologyEntity, Identifier],
            Iterable[Union[OntologyEntity, Identifier]],
        ],
    ):
        """Remove an ontology entity from the session."""
        entities = frozenset(
            entity
            for x in entities
            for entity in (
                x if not isinstance(x, (OntologyEntity, Identifier)) else (x,)
            )
        )

        for entity in entities:
            if isinstance(entity, OntologyEntity) and entity not in self:
                raise ValueError(f"Entity {entity} not contained in {self}.")

        for entity in entities:
            if isinstance(entity, OntologyEntity):
                entity = entity.identifier
            self._track_identifiers(entity, delete=True)
            self._graph.remove((entity, None, None))
            self._graph.remove((None, None, entity))

    def clear(self, force: bool = False):
        """Clear all the data stored in the session."""
        graph = self._graph_writable if force else self._graph
        graph.remove((None, None, None))
        self._namespaces.clear()
        self.entity_cache_timestamp = datetime.now()
        self.from_identifier.cache_clear()
        self.from_label.cache_clear()

        # Reload the essential TBox required by ontologies.
        if self.ontology is self:
            for parser in (
                OntologyParser.get_parser("simphony"),
                OntologyParser.get_parser("owl"),
                OntologyParser.get_parser("rdfs"),
            ):
                self.ontology.load_parser(parser)

    def get(
        self,
        *individuals: Union[OntologyIndividual, Identifier, str],
        oclass: Optional[OntologyClass] = None,
    ) -> Union[
        Set[OntologyIndividual],
        Optional[OntologyIndividual],
        Tuple[Optional[OntologyIndividual]],
    ]:
        """Return the individuals in the session.

        The structure of the output can vary depending on the form used for
        the call. See the "Returns:" section of this
        docstring for more details on this.

        Args:
            individuals: Restrict the individuals to be returned to a certain
                subset of the individuals in the session.
            oclass: Only yield ontology individuals which belong to a subclass
                of the given ontology class. Defaults to None (no filter).

        Returns:
            Calls without `*individuals` (SessionSet): The result of the
                call is a set-like object. This corresponds to
                the calls `get()`, `get(oclass=___)`.
            Calls with `*individuals` (Optional[OntologyIndividual],
                    Tuple[Optional["OntologyIndividual"], ...]):
                The position of each element in the result is determined by the
                position of the corresponding identifier/individual in the
                given list of identifiers/individuals. In this case, the result
                can contain `None` values if a given identifier/individual is
                not in the session, or if it does not satisfy the class
                filter. This description corresponds to the calls
                `get(*individuals)`, `get(*individuals, oclass=`___`)`.

        Raises:
            TypeError: Objects that are not ontology individuals,
                identifiers or strings provided as positional arguments.
            TypeError: Object that is not an ontology class passed as
                keyword argument `oclass`.
            RuntimeError: Ontology individuals that belong to a different
                session provided.
        """
        identifiers = list(individuals)
        for i, x in enumerate(identifiers):
            if not isinstance(x, (OntologyIndividual, Identifier, str)):
                raise TypeError(
                    f"Expected {OntologyIndividual}, {Identifier} or {str} "
                    f"objects, not {type(x)}."
                )
            elif isinstance(x, OntologyIndividual) and x not in self:
                raise RuntimeError(
                    "Cannot get an individual that belongs to "
                    "a different session."
                )

            if isinstance(x, str):
                if not isinstance(x, Identifier):
                    identifiers[i] = URIRef(x)
            elif isinstance(x, OntologyIndividual):
                identifiers[i] = x.identifier

        if identifiers:
            entities = [None] * len(identifiers)
            for i, identifier in enumerate(identifiers):
                try:
                    entity = self.from_identifier(identifier)
                except KeyError:
                    entity = None
                if entity and oclass and not entity.is_a(oclass):
                    entity = None
                entities[i] = entity

            if len(identifiers) == 1:
                entities = entities[0]
            else:
                entities = tuple(entities)
        else:
            entities = SessionSet(session=self, oclass=oclass)

        return entities

    def iter(
        self,
        *individuals: Union[OntologyIndividual, Identifier, str],
        oclass: Optional[OntologyClass] = None,
    ) -> Union[
        Iterator[OntologyIndividual],
        Iterator[Optional[OntologyIndividual]],
    ]:
        """Iterate over the ontology individuals in the session.

        The structure of the output can vary depending on the form used for
        the call. See the "Returns:" section of this docstring for more
        details on this.

        Args:
            individuals: Restrict the individuals to be returned to a certain
                subset of the individuals in the session.
            oclass: Only yield ontology individuals which belong to a subclass
                of the given ontology class. Defaults to None (no filter).

        Returns:
            Calls without `*individuals` (Iterator[OntologyIndividual]): The
                position of each element in the result is non-deterministic.
                This corresponds to the calls `iter()`, `iter(oclass=___)`.
            Calls with `*individuals` (Iterator[Optional[
                    OntologyIndividual]]):
                The position of each element in the result is determined by the
                position of the corresponding identifier/individual in the
                given list of identifiers/individuals. In this case, the result
                can contain `None` values if a given identifier/individual is
                not in the session, or if it does not satisfy the class
                filter. This description corresponds to the calls
                `iter(*individuals)`, `iter(*individuals, oclass=`___`)`.

        Raises:
            TypeError: Objects that are not ontology individuals,
                identifiers or strings provided as positional arguments.
            TypeError: Object that is not an ontology class passed as
                keyword argument `oclass`.
            RuntimeError: Ontology individuals that belong to a different
                session provided.
        """
        identifiers = list(individuals)
        for i, x in enumerate(identifiers):
            if not isinstance(x, (OntologyIndividual, Identifier, str)):
                raise TypeError(
                    f"Expected {OntologyIndividual}, {Identifier} or {str} "
                    f"objects, not {type(x)}."
                )
            elif isinstance(x, OntologyIndividual) and x not in self:
                raise RuntimeError(
                    "Cannot get an individual that belongs to "
                    "a different session."
                )

            if isinstance(x, str):
                if not isinstance(x, Identifier):
                    identifiers[i] = URIRef(x)
            elif isinstance(x, OntologyIndividual):
                identifiers[i] = x.identifier

        if oclass is not None and not isinstance(oclass, OntologyClass):
            raise TypeError(
                "Found object of type %s passed to argument "
                "oclass. Should be an OntologyClass." % type(oclass)
            )

        if identifiers:

            # The yield statement is encapsulated inside a function so that the
            # main function uses the return statement instead of yield. In this
            # way, exceptions are checked when the `iter` method is called
            # instead of when asking for the first result.
            def iterator() -> Iterator[Optional[OntologyIndividual]]:
                for identifier in identifiers:
                    try:
                        entity = self.from_identifier(identifier)
                    except KeyError:
                        entity = None
                    if entity and oclass and not entity.is_a(oclass):
                        entity = None
                    yield entity

            return iterator()

        else:
            return iter(SessionSet(session=self, oclass=oclass))

    # ↑ --------------------- Public API --------------------- ↑ #

    default_ontology: Session
    """The default ontology.

    When no T-Box is explicitly assigned to a session, this is the ontology
    it makes use of.
    """

    entity_cache_timestamp: Optional[datetime] = None
    """A timestamp marking the time when the session's graph was last modified.

    This timestamp is used by `OntologyEntity` and its subclasses to know
    whether they should invalidate their cache (e.g. the cache of the
    `superclasses` method must be invalidated when the session is cleared or a
    new ontology is loaded into the session).
    """

    _ontology: Optional[Session] = None
    """Private pointer to the T-Box of the session.

    Not `None` only when the T-Box of the session should be different from
    the default T-Box (the one referred to by the attribute `default_ontology`,
    which is by default a session containing all the installed ontologies).
    """

    def __init__(
        self,
        base: Optional[Graph] = None,  # The graph must be OPEN already.
        driver: Optional[InterfaceDriver] = None,
        ontology: Optional[Union[Session, bool]] = None,
        identifier: Optional[str] = None,
        namespaces: Dict[str, URIRef] = None,
        from_parser: Optional[OntologyParser] = None,
    ):
        """Initialize the session."""
        super().__init__()
        self._environment_references.add(self)
        # Base the session graph either on a store if passed or an empty graph.
        if base is not None:
            self._graph_writable = base
            self._graph = base

        else:
            graph = Graph()
            self._graph_writable = graph
            self._graph = graph

        self._interface_driver = driver

        # Configure the ontology for this session
        if isinstance(ontology, Session):
            self.ontology = ontology
        elif ontology is True:
            self._graph = ReadOnlyGraphAggregate([self._graph_writable])
            self.ontology = self
        elif ontology is not None:
            raise TypeError(
                f"Invalid ontology argument: {ontology}."
                f"Expected either a {Session} or {bool} object, "
                f"got {type(ontology)} instead."
            )

        if self.ontology is not self:
            """Bypass cache if this session is not a T-Box"""

            def bypass_cache(method: Callable):
                wrapped_func = method.__wrapped__

                @wraps(wrapped_func)
                def bypassed(*args, **kwargs):
                    return wrapped_func(self, *args, **kwargs)

                bypassed.cache_clear = lambda: None
                return bypassed

            self.from_identifier = bypass_cache(self.from_identifier)
            self.from_label = bypass_cache(self.from_label)
        else:
            """Log the time of last entity cache clearing."""

            self.entity_cache_timestamp = datetime.now()

        self._entity_cache = dict()

        self.creation_set = set()
        self._storing = list()

        self._namespaces = dict()
        # Load the essential TBox required by ontologies.
        if self.ontology is self:
            for parser in (
                OntologyParser.get_parser("simphony"),
                OntologyParser.get_parser("owl"),
                OntologyParser.get_parser("rdfs"),
            ):
                self.ontology.load_parser(parser)
        if from_parser:  # Compute session graph from an ontology parser.
            if self.ontology is not self:
                raise RuntimeError(
                    "Cannot load parsers in sessions which "
                    "are not their own ontology. Load the "
                    "parser on the ontology instead."
                )
            if namespaces is not None:
                logger.warning(
                    f"Namespaces bindings {namespaces} ignored, "
                    f"as the session {self} is being created from "
                    f"a parser."
                )
            self.load_parser(from_parser)
            self.identifier = identifier or from_parser.identifier
        else:  # Create an empty session.
            self.identifier = identifier
            namespaces = namespaces if namespaces is not None else dict()
            for key, value in namespaces.items():
                self.bind(key, value)

    def merge(self, entity: OntologyEntity) -> None:
        """Merge a given ontology entity with what is in the session.

        Copies the ontology entity to the session, but does not remove any
        old triples referring to the entity.

        Args:
            entity: The ontology entity to store.
        """
        self._update_and_merge_helper(entity, mode=False)

    def update(self, entity: OntologyEntity) -> None:
        """Store a copy of given ontology entity in the session.

        Args:
            entity: The ontology entity to store.
        """
        self._update_and_merge_helper(entity, mode=True)

    @property
    def namespaces(self) -> List[OntologyNamespace]:
        """Get all the namespaces bound to the session."""
        return [
            OntologyNamespace(iri=iri, name=name, ontology=self.ontology)
            for iri, name in self.ontology._namespaces.items()
        ]

    def bind(self, name: Optional[str], iri: Union[str, URIRef]):
        """Bind a namespace to this session.

        Args:
            name: the name to bind. The name is optional, a namespace object
                can be bound without name.
            iri: the IRI of the namespace to be bound to such name.
        """
        iri = URIRef(iri)
        for key, value in self.ontology._namespaces.items():
            if value == name and key != iri:
                raise ValueError(
                    f"Namespace {key} is already bound to name "
                    f"{name} in ontology {self}. "
                    f"Please unbind it first."
                )
        else:
            self.ontology._namespaces[iri] = name
            self.ontology._graph_writable.bind(name, iri)

    def unbind(self, name: Union[str, URIRef]):
        """Unbind a namespace from this session.

        Args:
            name: the name to which the namespace is already bound, or the
                IRI of the namespace.
        """
        for key, value in dict(self.ontology._namespaces).items():
            if value == name or key == URIRef(name):
                del self.ontology._namespaces[key]

    def get_namespace_bind(
        self, namespace: Union[OntologyNamespace, URIRef, str]
    ) -> Optional[str]:
        """Returns the name used to bind a namespace to the ontology.

        Args:
            namespace: Either an OntologyNamespace or the IRI of a namespace.

        Raises:
            KeyError: Namespace not bound to to the ontology.
        """
        ontology = self.ontology
        if isinstance(namespace, OntologyNamespace):
            ontology = namespace.ontology
            namespace = namespace.iri
        else:
            namespace = URIRef(namespace)

        not_bound_error = KeyError(
            f"Namespace {namespace} not bound to " f"ontology {self}."
        )
        if ontology is not self.ontology:
            raise not_bound_error
        try:
            return self.ontology._namespaces[namespace]
        except KeyError:
            raise not_bound_error

    def get_namespace(self, name: Union[str, URIRef]) -> OntologyNamespace:
        """Get a namespace registered with the session.

        Args:
            name: The namespace name or IRI to search for.

        Returns:
            The ontology namespace.

        Raises:
            KeyError: Namespace not found.
        """
        coincidences = iter(tuple())
        if isinstance(name, URIRef):
            coincidences_iri = (x for x in self.namespaces if x.iri == name)
            coincidences = itertools.chain(coincidences, coincidences_iri)
        elif isinstance(name, str):
            coincidences_name = (x for x in self.namespaces if x.name == name)
            coincidences = itertools.chain(coincidences, coincidences_name)
            # Last resort: user provided string but may be an IRI.
            coincidences_fallback = (
                x for x in self.namespaces if x.iri == URIRef(name)
            )
            coincidences = itertools.chain(coincidences, coincidences_fallback)

        result = next(coincidences, None)
        if result is None:
            raise KeyError(f"Namespace {name} not found in ontology {self}.")
        return result

    @property
    def graph(self) -> Graph:
        """Returns the session's graph."""
        return self._graph

    @property
    def driver(self) -> Optional[InterfaceDriver]:
        """The SimPhoNy interface on which the base graph is based on.

        Points to the interface response for realizing the base graph of the
        session. Not all graphs have to be based on an interface. In such
        cases, the value of this attribute is `None`.
        """
        return self._interface_driver

    @classmethod
    def get_default_session(cls) -> Optional[Session]:
        """Returns the default session."""
        for environment in cls._stack_default_environment[::-1]:
            if isinstance(environment, Session):
                return environment
        else:
            return None

    @classmethod
    def set_default_session(cls, session: Session):
        """Sets the first session of the stack of sessions.

        This effectively makes it the default. The method will not work if
        there are any other default environments in the stack
        """
        if len(cls._stack_default_environment) > 1:
            raise RuntimeError(
                "The default session cannot be changed when "
                "there are other environments in the stack."
            )
        try:
            cls._stack_default_environment.pop()
        except IndexError:
            pass
        cls._stack_default_environment.append(session)

    def load_parser(self, parser: OntologyParser):
        """Merge ontology packages with this ontology from a parser object.

        Args:
            parser: the ontology parser from where to load the new namespaces.
        """
        if self.ontology is not self:
            raise ModificationException()
        self._graph_writable += parser.graph
        for name, iri in parser.namespaces.items():
            self.bind(name, iri)
        self.entity_cache_timestamp = datetime.now()
        self.from_identifier.cache_clear()
        self.from_label.cache_clear()

    def iter_identifiers(self) -> Iterator[Union[BNode, URIRef]]:
        """Iterate over all the ontology entity identifiers in the session."""
        # Warning: identifiers can be repeated.
        supported_entity_types = frozenset(
            {
                # owl:AnnotationProperty
                OWL.AnnotationProperty,
                RDF.Property,
                # owl:DatatypeProperty
                OWL.DatatypeProperty,
                # owl:ObjectProperty
                OWL.ObjectProperty,
                # owl:Class
                OWL.Class,
                RDFS.Class,
                # owl:Restriction
                OWL.Restriction,
            }
        )

        # Yield the entities from the TBox (literals filtered out above).
        if self.ontology is self:
            yield from (
                s
                for t in supported_entity_types
                for s in self.ontology.graph.subjects(RDF.type, t)
                if not isinstance(s, Literal)
            )

        # Yield the entities from the ABox (literals filtered out below).
        yield from (
            t[0]
            for t in self._graph.triples((None, RDF.type, None))
            if not isinstance(t[0], Literal)
            and any(
                (t[2], RDF.type, supported_entity_type) in self.ontology.graph
                for supported_entity_type in supported_entity_types
            )
        )

    def iter_labels(
        self,
        entity: Optional[Union[Identifier, OntologyEntity]] = None,
        lang: Optional[str] = None,
        return_prop: bool = False,
        return_literal: bool = True,
        return_identifier: bool = False,
    ) -> Iterator[
        Union[
            Literal,
            str,
            Tuple[Union[Literal, str], Node],
            Tuple[Union[Literal, str], Node, Node],
        ]
    ]:
        """Iterate over all the labels of the entities in the session."""
        from simphony_osp.ontology.entity import OntologyEntity

        if isinstance(entity, OntologyEntity):
            entity = entity.identifier

        def filter_language(literal):
            if lang is None:
                return True
            elif lang == "":
                return literal.language is None
            else:
                return literal.language == lang

        labels = (
            (prop, literal, subject)
            for prop in self.label_properties
            for subject, _, literal in self._graph.triples(
                (entity, prop, None)
            )
        )
        labels = filter(
            lambda label_tuple: filter_language(label_tuple[1]), labels
        )
        if not return_prop and not return_literal and not return_identifier:
            return (str(x[1]) for x in labels)
        elif return_prop and not return_literal and not return_identifier:
            return ((str(x[1]), x[0]) for x in labels)
        elif not return_prop and return_literal and not return_identifier:
            return (x[1] for x in labels)
        elif return_prop and return_literal and not return_identifier:
            return ((x[1], x[0]) for x in labels)
        elif not return_prop and not return_literal and return_identifier:
            return ((str(x[1]), x[2]) for x in labels)
        elif return_prop and not return_literal and return_identifier:
            return ((str(x[1]), x[0], x[2]) for x in labels)
        elif not return_prop and return_literal and return_identifier:
            return ((x[1], x[2]) for x in labels)
        else:  # everything true
            return ((x[1], x[0], x[2]) for x in labels)

    def get_identifiers(self) -> Set[Identifier]:
        """Get all the identifiers in the session."""
        return set(self.iter_identifiers())

    def get_entities(self) -> Set[OntologyEntity]:
        """Get all the entities stored in the session."""
        return {x for x in self}

    _interface_driver: Optional[InterfaceDriver] = None

    def _update_and_merge_helper(
        self,
        entity: OntologyEntity,
        mode: bool,
        visited: Optional[set] = None,
    ) -> None:
        """Private `merge` and `update` helper.

        Args:
            entity: The ontology entity to merge.
            mode: True means update, False means merge.
            visited: Entities that have already been updated or merged.
        """
        if entity.session is None:  # Newly created entity.
            if mode:
                self._graph.remove((entity.iri, None, None))
            for t in entity.graph.triples((None, None, None)):
                self._graph.add(t)
        elif entity not in self:  # Entity from another session.
            active_relationship = self.ontology.from_identifier(
                simphony_namespace.activeRelationship
            )
            try:
                existing = self.from_identifier(entity.identifier)
            except KeyError:
                existing = None

            self._track_identifiers(entity.identifier)

            # Clear old types, active relationships and attributes for
            # the update operation.
            if existing and mode:
                # Clear old types.
                self._graph.remove((existing.identifier, RDF.type, None))

                for p in existing.graph.predicates(existing.identifier, None):
                    try:
                        predicate = self.ontology.from_identifier(p)
                    except KeyError:
                        continue
                    # Clear attributes or active relationships.
                    if isinstance(predicate, OntologyAttribute) or (
                        isinstance(predicate, OntologyRelationship)
                        and active_relationship in predicate.superclasses
                    ):
                        self._graph.remove((existing.identifier, p, None))

            # Merge new types, active relationships and attributes.
            # The double for loop pattern (first loop over p, then loop over o)
            # is used because calling `self.ontology.from_identifier` is
            # expensive.
            visited = visited if visited is not None else set()
            for p in entity.graph.predicates(entity.identifier, None):
                try:
                    predicate = self.ontology.from_identifier(p)
                except KeyError:
                    # Merge new types.
                    if p == RDF.type:
                        for o in entity.graph.objects(entity.identifier, p):
                            self._graph.add((entity.identifier, p, o))
                    continue

                for o in entity.graph.objects(entity.identifier, p):
                    if isinstance(predicate, OntologyAttribute):
                        # Merge attributes.
                        self._graph.add((entity.identifier, p, o))
                    elif (
                        isinstance(predicate, OntologyRelationship)
                        and active_relationship in predicate.superclasses
                    ):
                        # Merge active relationships.
                        obj = entity.session.from_identifier(o)
                        if not isinstance(obj, OntologyIndividual):
                            continue
                        if obj.identifier not in visited:
                            visited.add(obj.identifier)
                            self._update_and_merge_helper(obj, mode, visited)
                        self._graph.add((entity.identifier, p, o))

    creation_set: Set[Identifier]
    _namespaces: Dict[URIRef, str]
    _graph: Graph
    _graph_writable: Graph
    _driver: Optional[Interface] = None

    @property
    def _session_linked(self) -> Session:
        return self

    def _track_identifiers(self, identifier, delete=False):
        # Keep track of new additions while inside context manager.
        if delete:
            self.creation_set.discard(identifier)
        else:
            entity_triples_exist = (
                next(self._graph.triples((identifier, None, None)), None)
                is not None
            )
            if not entity_triples_exist:
                self.creation_set.add(identifier)


class QueryResult(SPARQLResult):
    """SPARQL query result."""

    session: Session

    def __init__(self, *args, session: Optional[Session] = None, **kwargs):
        """Initialize the query result.

        Namely, a session is linked to this query result so that if ontology
        individuals are requested,

        Args:
            session: Session to which this result is linked to.
        """
        self.session = session or Session.get_default_session()
        super().__init__(*args, **kwargs)

    # ↓ --------------------- Public API --------------------- ↓ #

    def __call__(
        self, **kwargs
    ) -> Union[Iterator[Triple], Iterator[bool], Iterator[ResultRow]]:
        """Select the datatypes of the query results ofr SELECT queries.

        Args:
            **kwargs: For each variable name on the query, a callable
                can be specified as keyword argument. When retrieving
                results, this callable will be run on the RDFLib item from the
                result. Literals are an exception. The callable will
                be applied on top of the result of the `toPython()` method
                of the callable.

        Raises:
            ValueError: When the query that produced this result object is
                not a SELECT query.
        """
        if self.type != "SELECT":
            if kwargs:
                raise ValueError(
                    f"Result datatypes cannot be converted for "
                    f"{self.type} queries."
                )
            yield from self
            return

        for key, value in kwargs.items():
            """Filter certain provided callables and replace them by others."""
            if isclass(value) and issubclass(value, OntologyIndividual):
                """Replace OntologyIndividual with spawning the individual
                from its identifier."""
                kwargs[key] = lambda x: self.session.from_identifier_typed(
                    x, typing=OntologyIndividual
                )

        for row in self:
            """Yield the rows with the applied datatype transformation."""
            values = {
                Variable(var): kwargs.get(str(var), lambda x: x)(
                    row[i]
                    if not isinstance(row[i], Literal)
                    else row[i].toPython()
                )
                for i, var in enumerate(self.vars)
            }
            yield ResultRow(values, self.vars)

    # ↑ --------------------- Public API --------------------- ↑ #


Session.default_ontology = Session(
    identifier="default ontology", ontology=True
)
Session.set_default_session(Session(identifier="default session"))
# This default ontology is later overwritten by simphony_osp/utils/pico.py



"""Developer tools."""

from simphony_osp.interfaces.interface import Interface as Wrapper
from simphony_osp.interfaces.remote.common import get_hash
from simphony_osp.ontology.operations import Operations, find_operations

__all__ = ["Operations", "Wrapper", "find_operations", "get_hash"]



"""An ontology individual."""
from __future__ import annotations

import functools
import logging
from abc import ABC
from itertools import chain
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    Mapping,
    MutableSet,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

from rdflib import OWL, RDF, Literal, URIRef
from rdflib.term import Identifier, Node

from simphony_osp.ontology.annotation import OntologyAnnotation
from simphony_osp.ontology.attribute import OntologyAttribute
from simphony_osp.ontology.entity import OntologyEntity
from simphony_osp.ontology.oclass import OntologyClass
from simphony_osp.ontology.operations.operations import OperationsNamespace
from simphony_osp.ontology.relationship import OntologyRelationship
from simphony_osp.ontology.utils import DataStructureSet
from simphony_osp.utils.datatypes import (
    ATTRIBUTE_VALUE_TYPES,
    UID,
    AnnotationValue,
    AttributeValue,
    OntologyPredicate,
    PredicateValue,
    RelationshipValue,
    Triple,
)
from simphony_osp.utils.simphony_namespace import simphony_namespace

if TYPE_CHECKING:
    from simphony_osp.ontology.operations.container import ContainerEnvironment
    from simphony_osp.session.session import Session

logger = logging.getLogger(__name__)

RDF_type = RDF.type
OWL_NamedIndividual = OWL.NamedIndividual


class ResultEmptyError(Exception):
    """The result is unexpectedly empty."""


class MultipleResultsError(Exception):
    """Only a single result is expected, but there were multiple."""


class ExistingIndividualException(ValueError):
    """To be raised when a provided CUDS is already linked."""

    pass


class ObjectSet(DataStructureSet, ABC):
    """A set interface to an ontology individual's neighbors.

    This class looks like and acts like the standard `set`, but it
    is a template to implement classes that use either the attribute
    interface or the methods `relationships_connect`,
    `relationships_disconnect` and `relationships_iter` from
    the ontology individual.

    When an instance is read or when it is modified in-place,
    the interfaced methods are used to reflect the changes.

    This class does not hold any object-related information itself, thus
    it is safe to spawn multiple instances linked to the same property
    and ontology individual (when single-threading).
    """

    _predicate: Optional[OntologyPredicate]
    """Main predicate to which this object refers. It will be used
    whenever there is ambiguity on which predicate to use. Can be set to
    None, usually meaning all predicates (see the specific
    implementations of this class: `AttributeSet`,
    `RelationshipSet` and `AnnotationSet`)."""

    _individual: OntologyIndividual
    """The ontology individual to which this object is linked to.
    Whenever the set is modified, the modification will affect this ontology
    individual."""

    # Public API
    # ↓ ------ ↓

    @property
    def individual(self) -> OntologyIndividual:
        """Ontology individual that this set refers to."""
        return self._individual

    @property
    def predicate(self) -> OntologyPredicate:
        """Predicate that this set refers to."""
        return self._predicate

    def __repr__(self) -> str:
        """Return repr(self)."""
        return (
            set(self).__repr__()
            + " <"
            + (f"{self._predicate} " if self._predicate is not None else "")
            + f"of ontology individual {self._individual}>"
        )

    def one(
        self,
    ) -> Union[AnnotationValue, AttributeValue, RelationshipValue]:
        """Return one element.

        Return one element if the set contains one element, else raise
        an exception.

        Returns:
            The only element contained in the set.

        Raises:
            ResultEmptyError: No elements in the set.
            MultipleResultsError: More than one element in the set.
        """
        iter_self = iter(self)
        first_element = next(iter_self, StopIteration)
        if first_element is StopIteration:
            raise ResultEmptyError(
                f"No elements attached to "
                f"{self._individual} through "
                f"{self._predicate}."
            )
        second_element = next(iter_self, StopIteration)
        if second_element is not StopIteration:
            raise MultipleResultsError(
                f"More than one element attached "
                f"to {self._individual} through "
                f"{self._predicate}."
            )
        return first_element

    def any(
        self,
    ) -> Optional[Union[AnnotationValue, AttributeValue, RelationshipValue]]:
        """Return any element of the set.

        Returns:
            Any element from the set if the set is not empty, else None.
        """
        return next(iter(self), None)

    def all(self) -> ObjectSet:
        """Return all elements from the set.

        Returns:
            All elements from the set, namely the set itself.
        """
        return self

    # ↑ ------ ↑
    # Public API

    @property
    def _predicates(
        self,
    ) -> Optional[
        Union[
            Set[OntologyAttribute],
            Set[OntologyRelationship],
            Set[OntologyAnnotation],
        ]
    ]:
        """All the predicates to which this instance refers to.

        Returns:
            Such predicates, or `None` if no main predicate is
            associated with this `ObjectSet`.
        """
        return set(
            self._predicate.subclasses if self._predicate is not None else None
        )

    def __init__(
        self,
        predicate: Optional[OntologyPredicate],
        individual: OntologyIndividual,
    ):
        """Fix the linked predicate and CUDS object."""
        self._individual = individual
        self._predicate = predicate
        super().__init__()


class AttributeSet(ObjectSet):
    """A set interface to an ontology individual's attributes.

    This class looks like and acts like the standard `set`, but it
    is an interface to the `attributes_add`, attributes_set`,
    `attributes_delete`, `attributes_value_contains` and
    `attributes_value_generator` methods.

    When an instance is read, the methods `attributes_value_generator`
    and `attributes_value_contains` are used to fetch the data. When it
    is modified in-place, the methods `attributes_add`, `attributes_set`,
    and `attributes_delete` are used to reflect the changes.

    This class does not hold any attribute-related information itself, thus
    it is safe to spawn multiple instances linked to the same attribute
    and ontology individual (when single-threading).
    """

    # Public API
    # ↓ ------ ↓

    def __iter__(self) -> Iterator[AttributeValue]:
        """The values assigned to the referred predicates.

        Such predicates are the main attribute and its subclasses.

        Returns:
            The mentioned values.
        """
        yielded: Set[AttributeValue] = set()
        for value in chain(
            *(
                self._individual.attributes_value_generator(
                    attribute=attribute
                )
                for attribute in self._predicates
            )
        ):
            if value not in yielded:
                yielded.add(value)
                yield value

    def __contains__(self, item: AttributeValue) -> bool:
        """Check whether a value is assigned to the attribute."""
        return any(
            self._individual.attributes_value_contains(attribute, item)
            for attribute in self._predicates
        )

    def update(self, other: Iterable[AttributeValue]) -> None:
        """Update the set with the union of itself and others."""
        underlying_set = set(self)
        added = set(other).difference(underlying_set)
        self._individual.attributes_add(self._predicate, added)

    def intersection_update(self, other: Iterable[AttributeValue]) -> None:
        """Update the set with the intersection of itself and another."""
        underlying_set = set(self)
        intersection = underlying_set.intersection(other)
        removed = underlying_set.difference(intersection)
        for attribute in self._predicates:
            self._individual.attributes_delete(attribute, removed)

    def difference_update(self, other: Iterable[AttributeValue]) -> None:
        """Remove all elements of another set from this set."""
        removed = set(self) & set(other)
        for attribute in self._predicates:
            self._individual.attributes_delete(attribute, removed)

    def symmetric_difference_update(self, other: Set[AttributeValue]) -> None:
        """Update set with the symmetric difference of it and another."""
        underlying_set = set(self)
        symmetric_difference = underlying_set.symmetric_difference(other)
        added = symmetric_difference.difference(underlying_set)
        self._individual.attributes_add(self._predicate, added)
        removed = underlying_set.difference(symmetric_difference)
        for attribute in self._predicates:
            self._individual.attributes_delete(attribute, removed)

    # ↑ ------ ↑
    # Public API

    _predicate: OntologyAttribute

    @property
    def _predicates(self) -> Set[OntologyAttribute]:
        """All the attributes to which this instance refers to.

        Returns:
            Such predicates are the subproperties of the main predicate, or
            if it is none, all the subproperties.
        """
        predicates = super()._predicates
        if predicates is None:
            predicates = set(self._individual.attributes_generator())
            # The code below is technically true, but makes no
            #  difference due to how `attributes_generator` is written.
            # predicates = set(chain(
            #    subclasses
            #    for attributes in
            #    self._individual.attributes_generator(_notify_read=True)
            #    for subclasses in attributes.subclasses
            # ))
        return predicates

    def __init__(
        self,
        attribute: Optional[OntologyAttribute],
        individual: OntologyIndividual,
    ):
        """Fix the liked OntologyAttribute and ontology individual."""
        super().__init__(attribute, individual)


class RelationshipSet(ObjectSet):
    """A set interface to an ontology individual's relationships.

    This class looks like and acts like the standard `set`, but it
    is an interface to the `relationships_connect`, `relationships_disconnect`
    and `relationships_iter` methods.

    When an instance is read, the method `relationships_iter` is used to fetch
    the data. When it is modified in-place, the methods
    `relationships_connect` and `relationships_disconnect` are used to
    reflect the changes.

    This class does not hold any relationship-related information itself,
    thus it is safe to spawn multiple instances linked to the same
    relationship and ontology individual (when single-threading).
    """

    @staticmethod
    def prevent_class_filtering(func):
        """Decorator breaking methods when class filtering is enabled."""

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if self._class_filter is not None:
                raise RuntimeError(
                    "Cannot edit a set with a class filter in-place."
                )
            return func(self, *args, **kwargs)

        return wrapper

    # Bind static method to use as decorator.
    prevent_class_filtering = prevent_class_filtering.__get__(object, None)

    # Public API
    # ↓ ------ ↓

    def __iter__(self) -> Iterator[OntologyIndividual]:
        """Iterate over individuals assigned to `self._predicates`.

        Note: no class filter.

        Returns:
            The mentioned underlying set.
        """
        individual = self._individual.identifier
        graph = self._individual.session.graph
        ontology = self._individual.session.ontology
        predicates = self._predicates

        # Get the predicate IRIs to be considered.
        predicates_direct = {predicate.identifier for predicate in predicates}
        predicates_inverse = {
            p.identifier
            for predicate in predicates
            for p in (predicate.inverse,)
            if p is not None
        }
        if self._inverse:
            predicates_direct, predicates_inverse = (
                predicates_inverse,
                predicates_direct,
            )

        # Get the identifiers of the individuals connected to
        # `self._individual` through the allowed predicates.
        connected = set()
        triples = graph.triples((individual, None, None))
        connected |= {o for s, p, o in triples if p in predicates_direct}
        triples = graph.triples((None, None, individual))
        connected |= {s for s, p, o in triples if p in predicates_inverse}
        identifiers = (
            tuple(uid.to_identifier() for uid in self._uid_filter)
            if self._uid_filter
            else tuple()
        )
        if identifiers:
            connected &= set(identifiers)
        if self._class_filter:
            connected &= {
                identifier
                for identifier in connected
                if self._class_filter
                in (
                    subclass
                    for c in graph.objects(identifier, RDF_type)
                    if c != OWL_NamedIndividual
                    for subclass in ontology.from_identifier_typed(
                        c, typing=OntologyClass
                    ).superclasses
                )
            }

        if self._uid_filter:
            yield from (
                self._individual.session.from_identifier_typed(
                    identifier, typing=OntologyIndividual
                )
                if identifier in connected
                else None
                for identifier in identifiers
            )
        else:
            yield from (
                self._individual.session.from_identifier_typed(
                    identifier, typing=OntologyIndividual
                )
                for identifier in connected
            )

    def __contains__(self, item: OntologyIndividual) -> bool:
        """Check if an individual is connected via the relationship."""
        if item not in self.individual.session:
            return False

        original_uid_filter = self._uid_filter
        try:
            self._uid_filter = (item.uid,)
            return next(iter(self)) is not None
        finally:
            self._uid_filter = original_uid_filter

    def __invert__(self) -> RelationshipSet:
        """Get the inverse RelationshipSet."""
        return self.inverse

    @property
    def inverse(self) -> RelationshipSet:
        """Get the inverse RelationshipSet.

        Returns a RelationshipSet that works in the inverse direction: the
        ontology individuals displayed are the ones which are the subject of
        the relationship.
        """
        return RelationshipSet(
            relationship=self._predicate,
            individual=self.individual,
            oclass=self._class_filter,
            inverse=not self._inverse,
        )

    @prevent_class_filtering
    def update(self, other: Iterable[OntologyIndividual]) -> None:
        """Update the set with the union of itself and other."""
        # The individuals to update might be already attached. Given an
        #  individual from `other`, several situations may arise:
        #
        #    1 - The relationship through which it is already attached is
        #        the same as the main predicate `self._predicate`. It is
        #        safe to attach it again, the same connection cannot be
        #        duplicated in the RDF standard.
        #
        #    2 - The relationship through which it is already attached is a
        #        sub-relationship of the main predicate. In such case,
        #        we keep the existing connection and do not add a new
        #        connection. The principle is: the more specific the
        #        knowledge is, the better.
        #
        #    3 - The relationship through which it is already attached is a
        #        super-relationship of the main predicate. Then it can make
        #        sense to remove the original connection and replace it
        #        with a new, more specific connection using the main
        #        predicate.
        #
        added = filter(lambda x: x not in self, other)  # Takes care of 2.
        # TODO: We do not take care of 3, because `.add` also does not
        #  take care of 3. This topic can be an object of discussion.
        for individual in added:
            self._connect(individual, rel=self._predicate)

    @prevent_class_filtering
    def intersection_update(self, other: Iterable[OntologyIndividual]) -> None:
        """Update the set with the intersection of itself and another."""
        # Note: please read the comment on the `update` method.
        underlying_set = set(self)
        result = underlying_set.intersection(other)

        removed = underlying_set.difference(result)
        if removed:
            for rel in self._predicates:
                self._disconnect(*removed, rel=rel)

        added = result.difference(underlying_set)
        self._connect(*added, rel=self._predicate)

    @prevent_class_filtering
    def difference_update(self, other: Iterable[OntologyIndividual]) -> None:
        """Remove all elements of another set from this set."""
        # Note: please read the comment on the `update` method.
        removed = set(self) & set(other)
        if removed:
            for rel in self._predicates:
                self._disconnect(*removed, rel=rel)

    @prevent_class_filtering
    def symmetric_difference_update(
        self, other: Iterable[OntologyIndividual]
    ) -> None:
        """Update with the symmetric difference of it and another."""
        # Note: please read the comment on the `update` method.
        underlying_set = set(self)
        result = underlying_set.symmetric_difference(other)

        removed = underlying_set.difference(result)
        if removed:
            for rel in self._predicates:
                self._disconnect(*removed, rel=rel)

        added = result.difference(underlying_set)
        self._connect(*added, rel=self._predicate)

    @prevent_class_filtering
    def _connect(
        self,
        *individuals: OntologyIndividual,
        rel: Optional[OntologyRelationship] = None,
    ):
        individuals = set(individuals)

        # Raise exception if any of the individuals to connect belongs to a
        # different session.
        different_session = {
            individual
            for individual in individuals
            if individual not in self.individual.session
        }
        if different_session:
            raise RuntimeError(
                f"Cannot connect ontology individuals belonging to a "
                f"different session: "
                f"{','.join(str(i) for i in different_session)}."
                f"Please add them to this session first using `session.add`."
            )

        rel = rel or self.predicate
        # Raise exception for predicates that are not a subclass of the
        # RelationshipSet's ontology predicates.
        if rel != self.predicate:
            allowed = (
                self._predicates if self.predicate is not None else {None}
            )
            if rel not in allowed:
                raise RuntimeError(
                    f"Predicate {rel} not within the set of allowed "
                    f"predicates {allowed}."
                )
        if rel is None:
            raise RuntimeError("No predicate specified.")
        rel = rel.identifier

        for individual in individuals:
            self.individual.session.graph.add(
                (self.individual.identifier, rel, individual.identifier)
            )

    @prevent_class_filtering
    def _disconnect(
        self,
        *individuals: OntologyIndividual,
        rel: Optional[OntologyRelationship] = None,
    ):
        individuals = set(individuals)

        # Raise exception if any of the individuals to connect belongs to a
        # different session.
        different_session = {
            individual
            for individual in individuals
            if individual not in self.individual.session
        }
        if different_session:
            raise RuntimeError(
                f"Cannot disconnect ontology individuals belonging to a "
                f"different session: "
                f"{','.join(str(i) for i in different_session)}."
                f"Please add them to this session first using `session.add`."
            )

        rel = rel or self.predicate
        # Raise exception for predicates that are not a subclass of the
        # RelationshipSet's ontology predicates.
        if rel != self.predicate:
            allowed = (
                self._predicates if self.predicate is not None else {None}
            )
            if rel not in allowed:
                raise RuntimeError(
                    f"Predicate {rel} not within the set of allowed "
                    f"predicates {allowed}"
                )
        if rel is None:
            raise RuntimeError("No predicate specified.")

        for rel in rel.subclasses:
            for individual in individuals:
                self.individual.session.graph.remove(
                    (
                        self.individual.identifier,
                        rel.identifier,
                        individual.identifier,
                    )
                )

    # ↑ ------ ↑
    # Public API

    _predicate: Optional[OntologyRelationship]
    _class_filter: Optional[OntologyClass]
    _uid_filter: Optional[Tuple[UID]]
    _inverse: bool = False

    def __init__(
        self,
        relationship: Optional[OntologyRelationship],
        individual: OntologyIndividual,
        oclass: Optional[OntologyClass] = None,
        inverse: bool = False,
        uids: Optional[Iterable[UID]] = None,
    ):
        """Fix the liked OntologyRelationship and ontology individual."""
        if relationship is not None and not isinstance(
            relationship, OntologyRelationship
        ):
            raise TypeError(
                "Found object of type %s. "
                "Should be an OntologyRelationship." % type(relationship)
            )
        if oclass is not None and not isinstance(oclass, OntologyClass):
            raise TypeError(
                "Found object of type %s. Should be "
                "an OntologyClass." % type(oclass)
            )
        uids = tuple(uids) if uids is not None else None
        if uids is not None:
            for uid in uids:
                if not isinstance(uid, UID):
                    raise TypeError(
                        "Found object of type %s. Should be an UID."
                        % type(uid)
                    )

        self._class_filter = oclass
        self._inverse = inverse
        self._uid_filter = uids
        super().__init__(relationship, individual)

    def iter_low_level(
        self,
    ) -> Union[
        Iterator[Tuple[Node, Optional[Node], Optional[bool]]],
        Iterator[Tuple[Node, Optional[Node], Optional[bool], Node]],
    ]:
        """Iterate over individuals assigned to `self._predicates`.

        Note: no class filter.

        Returns:
            The mentioned underlying set.
        """
        # Get the predicate IRIs (p) to be considered.
        # Let x be `self._individual`.
        #  - direct_allowed: Triples of the form (x, p, o) will result in o
        #    being a candidate to be yielded.
        #  - inverse_allowed: Triples of the form (s, p, x) will result in s
        #    being a candidate to be yielded.
        direct_allowed = {p.identifier for p in self._predicates}
        inverse_allowed = {
            rel.identifier
            for rel in filter(None, (p.inverse for p in self._predicates))
        }
        if self._inverse:
            direct_allowed, inverse_allowed = inverse_allowed, direct_allowed

        # Get the individuals connected to `self._individual` through the
        # allowed predicates, that is, o and s from the last comment.
        individual = self._individual.identifier
        graph = self._individual.session.graph
        if self._uid_filter is None:
            predicate_individual_direct = (
                (o, p)
                for p in direct_allowed
                for o in graph.objects(individual, p)
            )
            predicate_individual_inverse = (
                (s, p)
                for p in inverse_allowed
                for s in graph.subjects(p, individual)
            )
            individuals_and_relationships = chain(
                ((o, p, True) for o, p in predicate_individual_direct),
                ((s, p, False) for s, p in predicate_individual_inverse),
            )
        else:
            # In this case, we respect the ordering of the `_uid_filter` and
            # yield `(uid.to_identifier(), None, None)` when there is not an
            # allowed connection between `self._individual` the individual
            # represented by `uid`.
            def individuals_and_relationships():
                for uid in self._uid_filter:
                    identifier = uid.to_identifier()
                    found = chain(
                        (
                            (p, True)
                            for p in direct_allowed
                            if (individual, p, identifier) in graph
                        ),
                        (
                            (p, False)
                            for p in inverse_allowed
                            if (identifier, p, individual) in graph
                        ),
                    )
                    first = next(found, (None, None))
                    yield tuple((identifier, *first))
                    if first != (None, None):
                        yield from ((identifier, *f) for f in found)

            individuals_and_relationships = individuals_and_relationships()

        yield from individuals_and_relationships


class AnnotationSet(ObjectSet):
    """A set interface to an ontology individual's annotations.

    This class looks like and acts like the standard `set`, but it is an
    interface to the methods from `OntologyIndividual` that manage the
    annotations.

    This class does not hold any annotation-related information itself,
    thus it is safe to spawn multiple instances linked to the same
    relationship and ontology individual (when single-threading).
    """

    _predicate: OntologyAnnotation

    def __init__(
        self,
        annotation: Optional[OntologyAnnotation],
        individual: OntologyIndividual,
    ) -> None:
        """Fix the linked OntologyAnnotation and ontology individual."""
        super().__init__(annotation, individual)

    # Public API
    # ↓ ------ ↓

    def __iter__(self) -> Iterator[AnnotationValue]:
        """Iterate over annotations linked to the individual."""
        yield from self._individual.annotations_value_generator(
            annotation=self._predicate
        )

    def __contains__(self, item) -> bool:
        """Determine whether the individual is annotated with an item."""
        return super().__contains__(item)

    def update(self, other: Iterable[AnnotationValue]) -> None:
        """Update the set with the union of itself and other."""
        self._individual.annotations_add(
            annotation=self._predicate, values=other
        )

    def intersection_update(self, other: Iterable[AnnotationValue]) -> None:
        """Update the set with the intersection of itself and another."""
        self._individual.annotations_set(
            annotation=self._predicate, values=other
        )

    def difference_update(self, other: Iterable[Any]) -> None:
        """Return self-=other."""
        self._individual.annotations_delete(
            annotation=self._predicate, values=set(self) & set(other)
        )

    def symmetric_difference_update(
        self, other: Iterable[AnnotationValue]
    ) -> None:
        """Return self^=other."""
        self._individual.annotations_set(
            self._predicate, set(self) ^ set(other)
        )

    # ↑ ------ ↑
    # Public API


class OntologyIndividual(OntologyEntity):
    """An ontology individual."""

    rdf_identifier = Identifier

    def __init__(
        self,
        uid: Optional[UID] = None,
        session: Optional[Session] = None,
        triples: Optional[Iterable[Triple]] = None,
        merge: bool = False,
        class_: Optional[OntologyClass] = None,
        attributes: Optional[
            Mapping[OntologyAttribute, Iterable[AttributeValue]]
        ] = None,
    ) -> None:
        """Initialize the ontology individual."""
        if uid is None:
            uid = UID()
        elif not isinstance(uid, UID):
            raise Exception(
                f"Tried to initialize an ontology individual with "
                f"uid {uid}, which is not a UID object."
            )
        triples = set(triples) if triples is not None else set()
        # Attribute triples.
        attributes = attributes or dict()
        triples |= {
            (
                uid.to_iri(),
                k.iri,
                Literal(k.convert_to_datatype(e), datatype=k.datatype),
            )
            for k, v in attributes.items()
            for e in v
        }
        # Class triples.
        if class_:
            triples |= {(uid.to_iri(), RDF.type, class_.iri)}

        super().__init__(uid, session, triples or None, merge=merge)

    # Public API
    # ↓ ------ ↓

    @property
    def classes(self) -> FrozenSet[OntologyClass]:
        """Get the ontology classes of this ontology individual.

        Returns:
            A set with all the ontology classes of the ontology
            individual. When the individual has no classes, the set is empty.
        """
        return frozenset(
            self.session.ontology.from_identifier_typed(
                o, typing=OntologyClass
            )
            for o in self.session.graph.objects(self.identifier, RDF_type)
            if o != OWL_NamedIndividual
        )

    @classes.setter
    def classes(self, value: Iterable[OntologyClass]) -> None:
        """Set the ontology classes of this ontology individual.

        Args:
            value: New ontology classes of the ontology individual.
        """
        identifiers = set()
        for x in value:
            if not isinstance(x, OntologyClass):
                raise TypeError(f"Expected {OntologyClass}, not {type(x)}.")
            identifiers.add(x.identifier)
        self.session.graph.remove((self.identifier, RDF.type, None))
        for x in identifiers:
            self.session.graph.add((self.identifier, RDF.type, x))

    def is_a(self, ontology_class: OntologyClass) -> bool:
        """Check if the individual is an instance of the given ontology class.

        Args:
            ontology_class: The ontology class to test against.

        Returns:
            Whether the ontology individual is an instance of such ontology
                class.
        """
        return any(oc in ontology_class.subclasses for oc in self.classes)

    def __dir__(self) -> Iterable[str]:
        """Show the individual's attributes as autocompletion suggestions.

        This method takes care of the autocompletion for the dot notation.
        """
        attribute_autocompletion = filter(
            lambda x: x.isidentifier(), self._attribute_autocompletion()
        )
        return chain(super().__dir__(), attribute_autocompletion)

    def _ipython_key_completions_(self):
        """Show the individual's predicates as tab completion suggestions.

        The predicates include its attributes, relationships and annotations.

        This method is specific for the bracket notation ([], __getitem__).
        """
        return chain(
            super().__dir__(),
            self._attribute_autocompletion(),
            self._relationship_autocompletion(),
            self._annotation_autocompletion(),
        )

    def __getattr__(self, name: str) -> AttributeSet:
        """Retrieve the value of an attribute of the individual.

        Args:
            name: The label or suffix of the attribute.

        Raises:
            AttributeError: Unknown attribute label or suffix.
            AttributeError: Multiple attributes for the given label or suffix.

        Returns:
            The value of the attribute (a python object).
        """
        attributes = self._attributes_get_by_name(name)
        num_attributes = len(attributes)
        if num_attributes == 0:
            raise AttributeError(
                f"No attribute associated with {self} with label or prefix "
                f"{name} found."
            )
        elif num_attributes >= 2:
            error = (
                f"There are multiple attributes with label or suffix {name} "
                f"associated with {self}:"
                f" {', '.join(r.iri for r in attributes)}. "
                f"Please use an OntologyAttribute object together with the "
                f"indexing notation `individual[attribute]` to access the "
                f"values of this attribute."
            )
            raise AttributeError(error)
        attr = attributes.pop()

        values = self.attributes_value_generator(attr)
        value = next(values, None)
        if next(values, None) is not None:
            raise RuntimeError(
                f"Tried to fetch values of a "
                f"non-functional attribute {attr} using "
                f"the dot notation. This is not "
                f"supported. "
                f"\n \n"
                f"Please use subscript "
                f"notation instead for such attributes: "
                f"my_cuds[{attr}]. This will return a set "
                f"of values instead of a single one"
            )
        return value

    def __setattr__(
        self,
        name: str,
        value: Optional[Union[AttributeValue, Set[AttributeValue]]],
    ) -> None:
        """Set the value(s) of an attribute.

        Args:
            name: The label or suffix of the attribute.
            value: The new value.

        Raises:
            AttributeError: Unknown attribute label or suffix.
            AttributeError: Multiple attributes for the given label or suffix.
        """
        if name.startswith("_"):
            super().__setattr__(name, value)
            return

        try:
            attributes = self._attributes_get_by_name(name)
            num_attributes = len(attributes)
            if num_attributes == 0:
                raise AttributeError(
                    f"No attribute, associated with {self} with label or"
                    f"prefix {name} found."
                )
            elif num_attributes >= 2:
                error = (
                    f"There are multiple attributes with label or suffix "
                    f"{name} associated with {self}:"
                    f" {', '.join(r.iri for r in attributes)}. "
                    f"Please use an OntologyAttribute object together with the"
                    f" indexing notation `individual[attribute]` to access "
                    f"the values of this attribute."
                )
                raise AttributeError(error)
            attr = attributes.pop()
            value = {value} if value is not None else set()
            self.attributes_set(attr, value)
        except AttributeError as e:
            # Might still be an attribute of a subclass of OntologyIndividual.
            if hasattr(self, name):
                super().__setattr__(name, value)
            else:
                raise e

    def __getitem__(
        self, rel: Union[OntologyPredicate, str]
    ) -> Union[AttributeSet, RelationshipSet, AnnotationSet]:
        """Retrieve linked individuals, attribute values or annotation values.

        The subscripting syntax `individual[rel]` allows:
        - When `rel` is an OntologyAttribute, to obtain a set containing all
          the values assigned to the specified attribute. Such set can be
          modified in-place to change the assigned values.
        - When `rel` is an OntologyRelationship, to obtain a set containing
          all ontology individuals objects that are connected to `individual`
          through rel. Such set can be modified in-place to modify the
          existing connections.
        - When `rel` is an OntologyAnnotation, to obtain a set containing
          all the annotation values assigned to the specified annotation
          property. Such set can be modified in-place to modify the existing
          connections.
        - When `rel` is a string, the string is resolved to an
          OntologyAttribute, OntologyRelationship or OntologyAnnotation with a
          matching label, and then one of the cases above applies.

        The reason why a set is returned and not a list, or any other
        container allowing repeated elements, is that the underlying RDF
        graph does not accept duplicate statements.

        Args:
            rel: An ontology attribute, an ontology relationship or an ontology
                annotation (OWL datatype property, OWL object property,
                OWL annotation property). Alternatively a string, which will be
                resolved, using labels, to one of the classes described above.

        Raises:
            KeyError: Unknown attribute, relationship or annotation label or
                suffix.
            KeyError: Multiple attributes, relationships or annotations found
                for the given label or suffix.
            TypeError: Trying to use something that is neither an
                OntologyAttribute, an OntologyRelationship, an
                OntologyAnnotation or a string as index.
        """
        if isinstance(rel, str):
            """Resolve string to an attribute, relationship or annotation."""
            entities = set()

            # Try to find a matching attribute.
            try:
                entities |= self._attributes_get_by_name(rel)
            except AttributeError:
                pass

            # Try to find a matching relationship.
            entities |= {
                relationship
                for _, relationship in self.relationships_iter(return_rel=True)
                for label in relationship.iter_labels(return_literal=False)
                if rel == label or relationship.iri.endswith(rel)
            }

            # Try to find a matching annotation.
            entities |= {
                annotation
                for _, annotation in self.annotations_iter(return_rel=True)
                for label in annotation.iter_labels(
                    return_literal=False, return_prop=False
                )
                if rel == label or annotation.iri.endswith(rel)
            }

            num_entities = len(entities)
            if num_entities == 0:
                raise KeyError(
                    f"No attribute, relationship or annotation "
                    f"associated with {self} with label or suffix {rel} found."
                )
            elif num_entities >= 2:
                raise KeyError(
                    f"There are multiple attributes, relationships or "
                    f"annotations with label or suffix {rel} associated with "
                    f"{self}:"
                    f"{', '.join(r.iri for r in entities)}. "
                    f"Please use an OntologyAttribute, OntologyRelationship "
                    f"or OntologyAnnotation object together with the "
                    f"indexing notation `individual[entity]` to access the "
                    f"values of this attribute, relationship or annotation."
                )
            rel = entities.pop()

        if isinstance(rel, OntologyAttribute):
            set_class = AttributeSet
        elif isinstance(rel, OntologyRelationship):
            set_class = RelationshipSet
        elif isinstance(rel, OntologyAnnotation):
            set_class = AnnotationSet
        else:
            raise TypeError(
                f"Ontology individual indices must be ontology "
                f"relationships, ontology attributes, "
                f"or ontology annotations, not {type(rel)}."
            )
        return set_class(rel, self)

    def __setitem__(
        self,
        rel: Union[OntologyPredicate, str],
        values: Optional[Union[PredicateValue, Set[PredicateValue]]],
    ) -> None:
        """Manages object, data and annotation properties.

        The subscripting syntax `individual[rel] = ` allows,

        - When `rel` is an OntologyRelationship, to replace the list of
          ontology individuals that are connected to `individual` through rel.
        - When `rel` is an OntologyAttribute, to replace the values of
          such attribute.
        - When `rel` is an OntologyAnnotation, to replace the annotation
          values of such annotation property.
        - When `rel` is a string, the string is resolved to an
          OntologyAttribute, OntologyRelationship or OntologyAnnotation with a
          matching label, and then one of the cases above applies.

        This function only accepts hashable objects as input, as the
        underlying RDF graph does not accept duplicate statements.

        Args:
            rel: Either an ontology attribute, an ontology relationship or
                an ontology annotation (OWL datatype property, OWL object
                property, OWL annotation property). Alternatively a string,
                which will be resolved, using labels, to one of the classes
                described above.
            values: Either a single element compatible with the OWL standard
                (this includes ontology individuals) or a set of such
                elements.

        Raises:
            KeyError: Unknown attribute, relationship or annotation label or
                suffix.
            KeyError: Multiple attributes, relationships or annotations found
                for the given label or suffix.
            TypeError: Trying to assign attributes using an object property,
                trying to assign ontology individuals using a data property,
                trying to use something that is neither an OntologyAttribute,
                an OntologyRelationship, an OntologyAnnotation nor a string as
                index.
        """
        if (
            isinstance(values, ObjectSet)
            and values.individual is self
            and values.predicate is rel
        ):
            # Do not do anything when the set assigned is a set referring to
            #  self and referring to the same predicate that was specified.
            #  Avoids duplication of work that would happen because
            #  `x[c] += y <-> x[c] == x[c].__iadd__(y)`. An alternative is
            #  getting rid of `__iadd__` so that
            #  `x[c] += y <-> x[c] = x[c] + y`. But this implies
            #  incompatibilities with `collections.ABC` (they already define
            #  `__isub__` for MutableSet for example).
            return

        values = values or set()
        values = (
            {values} if not isinstance(values, (Set, MutableSet)) else values
        )
        # Apparently instances of MutableSet are not instances of Set.

        # Classify the values by type.
        values = self._classify_by_type(values)

        if isinstance(rel, str):
            """Resolve string to an attribute, relationship or annotation."""
            entities = set()

            # Try to find a matching attribute.
            try:
                entities |= self._attributes_get_by_name(rel)
            except AttributeError:
                pass

            # Try to find a matching relationship.
            entities |= {
                relationship
                for _, relationship in self.relationships_iter(return_rel=True)
                for label in relationship.iter_labels(return_literal=False)
                if rel == label or relationship.iri.endswith(rel)
            }

            # Try to find a matching annotation.
            entities |= {
                annotation
                for _, annotation in self.annotations_iter(return_rel=True)
                for label in annotation.iter_labels(
                    return_literal=False, return_prop=False
                )
                if rel == label or annotation.iri.endswith(rel)
            }

            num_entities = len(entities)
            if num_entities == 0:
                raise KeyError(
                    f"No attribute, relationship or annotation "
                    f"associated with {self} with label or suffix {rel} found."
                )
            elif num_entities >= 2:
                raise KeyError(
                    f"There are multiple attributes, relationships or "
                    f"annotations with label or suffix {rel} associated with "
                    f"{self}:"
                    f"{', '.join(r.iri for r in entities)}. "
                    f"Please use an OntologyAttribute, OntologyRelationship "
                    f"or OntologyAnnotation object together with the "
                    f"indexing notation `individual[entity]` to access the "
                    f"values of this attribute, relationship or annotation."
                )
            rel = entities.pop()

        # Perform assignments.
        if isinstance(rel, OntologyRelationship):
            if (len(values) > 0 and OntologyIndividual not in values) or len(
                values
            ) > 1:
                raise TypeError(
                    f"Trying to assign python objects which are "
                    f"not ontology individuals using an object "
                    f"property {rel}."
                )
            assigned = set(values.get(OntologyIndividual, set()))
            current = RelationshipSet(rel, self, oclass=None)

            add = assigned - current
            remove = current - assigned

            current -= remove
            current |= add
        elif isinstance(rel, OntologyAttribute):
            if (
                len(values) > 0
                and all(
                    x not in values for x in (ATTRIBUTE_VALUE_TYPES, Literal)
                )
            ) or len(values) > 2:
                raise TypeError(
                    f"Trying to assign python objects which "
                    f"cannot be interpreted as literals "
                    f"using a data property {rel}."
                )
            assigned = set(values.get(ATTRIBUTE_VALUE_TYPES, set())) | set(
                values.get(Literal, set())
            )
            current = AttributeSet(rel, self)

            add = assigned - current
            remove = current - assigned

            current |= add
            current -= remove
        elif isinstance(rel, OntologyAnnotation):
            # TODO: Use a unit of work pattern here like above to only
            #  remove and add, rather than replacing.
            self.annotations_set(rel, values)
        else:
            raise TypeError(
                f"Ontology individual indices must be ontology "
                f"relationships, ontology attributes or ontology "
                f"annotations not {type(rel)}."
            )

    def __delitem__(self, rel: OntologyPredicate):
        """Delete all objects attached through rel.

        Args:
            rel: Either an ontology attribute, an ontology relationship or
                an ontology annotation (OWL datatype property, OWL object
                property, OWL annotation property). Alternatively a string,
                which will be resolved, using labels, to one of the classes
                described above.
        """
        self.__setitem__(rel=rel, values=set())

    def connect(
        self,
        *individuals: Union[OntologyIndividual, Identifier, str],
        rel: Union[OntologyRelationship, Identifier],
    ) -> None:
        """Connect ontology individuals to other ontology individuals.

        Args:
            individuals: The objects to be added. Their identifiers may also
                be used.
            rel: The relationship between the objects.

        Raises:
            TypeError: Objects that are not ontology individuals,
                identifiers or strings provided as positional arguments.
            TypeError: Object that is not an ontology relationship or the
                identifier of an ontology relationship passed as keyword
                argument `rel`.
            RuntimeError: Ontology individuals that belong to a different
                session provided.
        """
        individuals = list(individuals)
        for i, x in enumerate(individuals):
            if isinstance(x, str):
                if not isinstance(x, Identifier):
                    x = URIRef(x)
                x = self.session.from_identifier_typed(
                    x, typing=OntologyIndividual
                )
                individuals[i] = x
            if not isinstance(x, OntologyIndividual):
                raise TypeError(
                    f"Expected {OntologyIndividual}, {Identifier} or {str} "
                    f"objects, not {type(x)}."
                )
        individuals = set(individuals)

        if isinstance(rel, Identifier):
            rel = self.session.ontology.from_identifier_typed(
                rel,
                typing=OntologyRelationship,
            )
        elif not isinstance(rel, OntologyRelationship):
            raise TypeError(
                "Found object of type %s passed to argument rel. "
                "Should be an OntologyRelationship." % type(rel)
            )

        self[rel] |= individuals

    def disconnect(
        self,
        *individuals: Union[OntologyIndividual, Identifier, str],
        rel: Union[OntologyRelationship, Identifier] = OWL.topObjectProperty,
        oclass: Optional[OntologyClass] = None,
    ) -> None:
        """Disconnect ontology individuals from this one.

        Args:
            individuals: Specify the individuals to disconnect. When no
                individuals are specified, all connected individuals are
                considered.
            rel: Only remove individuals which are connected by subclass of the
                given relationship. Defaults to OWL.topObjectProperty (any
                relationship).
            oclass: Only remove elements which are a subclass of the given
                ontology class. Defaults to None (no filter).

        Raises:
            TypeError: Objects that are not ontology individuals,
                identifiers or strings provided as positional arguments.
            TypeError: Object that is not an ontology relationship or the
                identifier of an ontology relationship passed as keyword
                argument `rel`.
            TypeError: Object that is not an ontology class passed as
                keyword argument `oclass`.
            RuntimeError: Ontology individuals that belong to a different
                session provided.
        """
        individuals = list(individuals)
        for i, x in enumerate(individuals):
            if isinstance(x, str):
                if not isinstance(x, Identifier):
                    x = URIRef(x)
                x = self.session.from_identifier_typed(
                    x, typing=OntologyIndividual
                )
                individuals[i] = x
            if not isinstance(x, OntologyIndividual):
                raise TypeError(
                    f"Expected {OntologyIndividual}, {Identifier} or {str} "
                    f"objects, not {type(x)}."
                )
        individuals = set(individuals)

        if isinstance(rel, Identifier):
            rel = self.session.ontology.from_identifier_typed(
                rel, typing=OntologyRelationship
            )
        elif not isinstance(rel, OntologyRelationship):
            raise TypeError(
                "Found object of type %s passed to argument rel. "
                "Should be an OntologyRelationship." % type(rel)
            )

        if oclass is not None and not isinstance(oclass, OntologyClass):
            raise TypeError(
                "Found object of type %s passed to argument "
                "oclass. Should be an OntologyClass." % type(oclass)
            )

        individuals = individuals or self[rel]

        if oclass:
            individuals = {x for x in individuals if x.is_a(oclass)}

        self[rel] -= individuals

    def get(
        self,
        *individuals: Union[OntologyIndividual, Identifier, str],
        rel: Union[OntologyRelationship, Identifier] = OWL.topObjectProperty,
        oclass: Optional[OntologyClass] = None,
        return_rel: bool = False,
    ) -> Union[
        RelationshipSet,
        Optional[OntologyIndividual],
        Tuple[Optional[OntologyIndividual], ...],
        Tuple[Tuple[OntologyIndividual, OntologyRelationship]],
    ]:
        """Return the connected individuals.

        The structure of the output can vary depending on the form used for
        the call. See the "Returns:" section of this
        docstring for more details on this.

        Args:
            individuals: Restrict the elements to be returned to a certain
                subset of the connected elements.
            rel: Only return individuals which are connected by a subclass
                of the given relationship. Defaults to
                OWL.topObjectProperty (any relationship).
            oclass: Only yield individuals which are a subclass of the given
                ontology class. Defaults to None (no filter).
            return_rel: Whether to return the connecting relationship.
                Defaults to False.

        Returns:
            Calls without `*individuals` (RelationshipSet): The result of the
                call is a set-like object. This corresponds to
                the calls `get()`, `get(rel=___)`, `get(oclass=___)`,
                `get(rel=___, oclass=___)`, with the parameter `return_rel`
                unset or set to False.
            Calls with `*individuals` (Optional[OntologyIndividual],
                    Tuple[Optional["OntologyIndividual"], ...]):
                The position of each element in the result is determined by
                the position of the corresponding identifier/individual in the
                given list of identifiers/individuals. In this case, the result
                can contain `None` values if a given identifier/individual is
                not connected to this individual, or if it does not satisfy
                the class filter. When only one individual or identifier is
                specified, a single object is returned instead of a Tuple.
                This description corresponds to the calls `get(*individuals)`,
                `get(*individuals, rel=___)`,
                `get(*individuals, rel=___, oclass=___)`, with the parameter
                `return_rel` unset or set to False.
            Calls with `return_rel=True` (Tuple[
                    Tuple[OntologyIndividual, OntologyRelationship]]):
                The dependence of the order of the elements is maintained
                for the calls with `*individuals`, a non-deterministic order is
                used for the calls without `*individuals`. No `None` values
                are contained in the result (such identifiers or individuals
                are simply skipped). Moreover, the elements returned are now
                pairs of individuals and the relationship connecting this
                individual to such ones. This description corresponds to any
                call of the form `get(..., return_rel=True)`.

        Raises:
            TypeError: Objects that are not ontology individuals,
                identifiers or strings provided as positional arguments.
            TypeError: Object that is not an ontology relationship or the
                identifier of an ontology relationship passed as keyword
                argument `rel`.
            TypeError: Object that is not an ontology class passed as
                keyword argument `oclass`.
            RuntimeError: Ontology individuals that belong to a different
                session provided.
        """
        identifiers = list(individuals)
        for i, x in enumerate(identifiers):
            if not isinstance(x, (OntologyIndividual, Identifier, str)):
                raise TypeError(
                    f"Expected {OntologyIndividual}, {Identifier} or {str} "
                    f"objects, not {type(x)}."
                )
            elif isinstance(x, OntologyIndividual) and x not in self.session:
                raise RuntimeError(
                    "Cannot get an individual that belongs to "
                    "a different session, please add it to this session "
                    "first using `session.add`."
                )

            if isinstance(x, str):
                if not isinstance(x, Identifier):
                    x = URIRef(x)
                identifiers[i] = UID(x)
            elif isinstance(x, OntologyIndividual):
                identifiers[i] = UID(x.identifier)

        if isinstance(rel, Identifier):
            rel = self.session.ontology.from_identifier_typed(
                rel, typing=OntologyRelationship
            )
        elif not isinstance(rel, OntologyRelationship):
            raise TypeError(
                "Found object of type %s passed to argument rel. "
                "Should be an OntologyRelationship." % type(rel)
            )

        if oclass is not None and not isinstance(oclass, OntologyClass):
            raise TypeError(
                "Found object of type %s passed to argument "
                "oclass. Should be an OntologyClass." % type(oclass)
            )

        relationship_set = RelationshipSet(
            rel, self, oclass=oclass, uids=tuple(identifiers) or None
        )

        if not return_rel:
            if not identifiers:
                return relationship_set
            else:
                return (
                    next(iter(relationship_set), None)
                    if len(identifiers) <= 1
                    else tuple(relationship_set)
                )
        else:
            result = []
            for (i, r, t) in relationship_set.iter_low_level():
                if not t:
                    continue
                session = self.session
                result += [
                    (
                        session.from_identifier_typed(
                            i, typing=OntologyIndividual
                        ),
                        session.ontology.from_identifier_typed(
                            r, typing=OntologyRelationship
                        ),
                    )
                ]
            return tuple(result)

    def iter(
        self,
        *individuals: Union[OntologyIndividual, Identifier, str],
        rel: Union[OntologyRelationship, Identifier] = OWL.topObjectProperty,
        oclass: Optional[OntologyClass] = None,
        return_rel: bool = False,
    ) -> Union[
        Iterator[OntologyIndividual],
        Iterator[Optional[OntologyIndividual]],
        Iterator[Tuple[OntologyIndividual, OntologyRelationship]],
    ]:
        """Iterate over the connected individuals.

        The structure of the output can vary depending on the form used for
        the call. See the "Returns:" section of this docstring for more
        details on this.

        Args:
            individuals: Restrict the elements to be returned to a certain
                subset of the connected elements.
            rel: Only yield individuals which are connected by a subclass
                of the given relationship. Defaults to
                OWL.topObjectProperty (any relationship).
            oclass: Only yield individuals which are a subclass of the given
                ontology class. Defaults to None (no filter).
            return_rel: Whether to yield the connecting relationship.
                Defaults to False.

        Returns:
            Calls without `*individuals` (Iterator[OntologyIndividual]): The
                position of each element in the result is non-deterministic.
                This corresponds to the calls `iter()`, `iter(rel=___)`,
                `iter(oclass=___)`, `iter(rel=___, oclass=___)`, with the
                parameter `return_rel` unset or set to False.
            Calls with `*individuals` (Iterator[Optional[
                    OntologyIndividual]]):
                The position of each element in the result is determined by the
                position of the corresponding identifier/individual in the
                given list of identifiers/individuals. In this case, the result
                can contain `None` values if a given identifier/individual is
                not connected to this individual, or if it does not satisfy
                the class filter. This description corresponds to the calls
                `iter(*individuals)`, `iter(*individuals, rel=___)`,
                `iter(*individuals, rel=___, oclass=`___`)`.
            Calls with `return_rel=True` (Iterator[
                    Tuple[OntologyIndividual, OntologyRelationship]]):
                The dependence of the order of the elements is maintained
                for the calls with `*individuals`. No `None` values
                are contained in the result (such identifiers or individuals
                are simply skipped). Moreover, the elements returned are now
                pairs of individualsand the relationship connecting this
                individual to such ones. This description corresponds to any
                call of the form `iter(..., return_rel=True)`.


        Raises:
            TypeError: Objects that are not ontology individuals,
                identifiers or strings provided as positional arguments.
            TypeError: Object that is not an ontology relationship or the
                identifier of an ontology relationship passed as keyword
                argument `rel`.
            TypeError: Object that is not an ontology class passed as
                keyword argument `oclass`.
            RuntimeError: Ontology individuals that belong to a different
                session provided.
        """
        identifiers = list(individuals)
        for n, x in enumerate(identifiers):
            if not isinstance(x, (OntologyIndividual, Identifier, str)):
                raise TypeError(
                    f"Expected {OntologyIndividual}, {Identifier} or {str} "
                    f"objects, not {type(x)}."
                )
            elif isinstance(x, OntologyIndividual) and x not in self.session:
                raise RuntimeError(
                    "Cannot get an individual that belongs to "
                    "a different session, please add it to this session "
                    "first using `session.add`."
                )

            identifiers[n] = UID(
                x.identifier if isinstance(x, OntologyIndividual) else x
            )

        if isinstance(rel, Identifier):
            rel = self.session.ontology.from_identifier_typed(
                rel, typing=OntologyRelationship
            )
        elif not isinstance(rel, OntologyRelationship):
            raise TypeError(
                "Found object of type %s passed to argument rel. "
                "Should be an OntologyRelationship." % type(rel)
            )

        if oclass is not None and not isinstance(oclass, OntologyClass):
            raise TypeError(
                "Found object of type %s passed to argument "
                "oclass. Should be an OntologyClass." % type(oclass)
            )

        relationship_set = RelationshipSet(
            rel, self, oclass=oclass, uids=tuple(identifiers) or None
        )

        # In the following section, return is used instead of yield so that
        # code runs until the `return` statement when iter is called (
        # otherwise the exceptions above would not be raised until the first
        # item from the iterator is requested).
        if not return_rel:
            iterator = iter(relationship_set)
        else:

            def iterator() -> Iterator[
                Tuple[OntologyIndividual, OntologyRelationship]
            ]:
                """Helper iterator.

                The purpose of defining this iterator is to be able to
                return it, instead of using the `yield` keyword on the main
                function, as described on the comment above.
                """
                for (i, r, t) in relationship_set.iter_low_level():
                    if not t:
                        continue
                    session = self.session
                    yield (
                        (
                            session.from_identifier_typed(
                                i, typing=OntologyIndividual
                            ),
                            session.ontology.from_identifier_typed(
                                r, typing=OntologyRelationship
                            ),
                        )
                    )

            iterator = iterator()
        return iterator

    @property
    def operations(self) -> OperationsNamespace:
        """Access operations specific this individual's class."""
        if self._operations_namespace is None:
            self._operations_namespace = OperationsNamespace(individual=self)
        return self._operations_namespace

    def __enter__(self) -> ContainerEnvironment:
        """Use an ontology individual as a context manager.

        At the moment, individuals cannot be used as a context manager,
        except for containers.

        Returns:
            A `ContainerEnvironment` object when a container is used as a
            context manager. `ContainerEnvironment` objects have the same
            functionality as the operations defined for a container,
            but those can be called directly.

        Raises:
            AttributeError: When an ontology individual which does not
                belong to the `Container` subclass is used as a context
                manager.
        """
        classes = {class_.identifier for class_ in self.superclasses}

        if simphony_namespace.Container in classes:
            # Triggers the creation of the operations instance and thus the
            # environment.
            environment = self.operations.environment
            self._operations_context = environment
            return environment.__enter__()

        raise AttributeError("__enter__")

    def __exit__(self, *args):
        """Leave the individual's context manager when used as such.

        At the moment, individuals cannot be used as a context manager,
        except for containers.

        Raises:
            AttributeError: When the method `__enter__` has not been called
                on the individual before.
        """
        if self._operations_context is not None:
            context_return = self._operations_context.__exit__(*args)
            self._operations_context = None
            return context_return

        raise AttributeError("__exit__")

    @property
    def attributes(
        self,
    ) -> Mapping[OntologyAttribute, FrozenSet[AttributeValue]]:
        """Get the attributes of this individual as a dictionary."""
        generator = self.attributes_attribute_and_value_generator()
        return MappingProxyType(
            {attr: frozenset(gen) for attr, gen in generator}
        )

    # ↑ ------ ↑
    # Public API

    _operations_namespace: Optional[OperationsNamespace] = None
    """Holds the operations namespace instance for this ontology individual.

    The namespace in turns holds the instances of any subclasses of
    `Operation` that were defined and compatible with the classes of the
    individual.
    """

    _operations_context: Optional[ContainerEnvironment] = None
    """Stores the current context object.

    Some individuals (currently only containers) can be used as context
    managers through the operations API. The way this works is the
    following: when the `with` statement is used, an operation on the
    individual is called that retrieves the actual context manager object
    (which is not an individual) and calls the `__enter__` method from it.

    This context manager object needs to be stored somewhere so that when
    the individual context manager is exited, the actual context manager
    object's `__exit__` method is also called. This is the purpose of this
    attribute.
    """

    def _attribute_autocompletion(self) -> Iterable[str]:
        """Compute individual's attributes as autocompletion suggestions."""
        result = iter(())
        attributes = (
            attr
            for oclass in self.classes
            for attr in chain(
                oclass.attributes.keys(),
                oclass.optional_attributes,
                self.attributes_generator(),
            )
        )
        for attribute in attributes:
            result = chain(
                result,
                attribute.iter_labels(return_literal=False),
                (attribute.iri[len(attribute.namespace.iri) :],),
            )
        return result

    def _relationship_autocompletion(self) -> Iterable[str]:
        """Compute individual's relationships as autocompletion suggestions."""
        result = iter(())
        relationships = (
            relationship
            for _, relationship in self.relationships_iter(return_rel=True)
        )
        for relationship in relationships:
            result = chain(
                result,
                relationship.iter_labels(return_literal=False),
                (relationship.iri[len(relationship.namespace.iri) :],),
            )
        return result

    def _annotation_autocompletion(self) -> Iterable[str]:
        """Compute individual's annotations as autocompletion suggestions."""
        result = iter(())
        annotation_properties = (
            annotation for annotation, _ in self.annotations_iter()
        )
        for annotation in annotation_properties:
            result = chain(
                result,
                annotation.iter_labels(
                    return_literal=False, return_prop=False
                ),
                (annotation.iri[len(annotation.namespace.iri) :],),
            )
        return result

    def _get_direct_superclasses(self) -> Iterable[OntologyClass]:
        yield from (
            x for oclass in self.classes for x in oclass.direct_superclasses
        )

    def _get_direct_subclasses(self) -> Iterable[OntologyClass]:
        yield from (
            x for oclass in self.classes for x in oclass.direct_subclasses
        )

    def _get_superclasses(self) -> Iterable[OntologyClass]:
        yield from (x for oclass in self.classes for x in oclass.superclasses)

    def _get_subclasses(self) -> Iterable[OntologyClass]:
        yield from (x for oclass in self.classes for x in oclass.subclasses)

    # Annotation handling
    # ↓ --------------- ↓

    @staticmethod
    def _classify_by_type(
        values: Set[PredicateValue],
    ) -> Dict[Type[PredicateValue], PredicateValue]:
        values = {
            type_: tuple(filter(lambda x: isinstance(x, type_), values))
            for type_ in (
                OntologyAnnotation,
                OntologyAttribute,
                OntologyClass,
                OntologyIndividual,
                OntologyRelationship,
                ATTRIBUTE_VALUE_TYPES,
                URIRef,
                Literal,
            )
        }
        values = {key: value for key, value in values.items() if value}
        return values

    def annotations_add(
        self,
        annotation: OntologyAnnotation,
        values: Union[
            Dict[Type[AnnotationValue], AnnotationValue], Set[AnnotationValue]
        ],
    ) -> None:
        """Adds annotations to the ontology individual."""
        if not isinstance(values, dict):
            values = self._classify_by_type(values)
        for value in chain(
            *(
                values.get(key, set())
                for key in (
                    OntologyAnnotation,
                    OntologyAttribute,
                    OntologyClass,
                    OntologyIndividual,
                    OntologyRelationship,
                )
            )
        ):
            self.session.graph.add((self.iri, annotation.iri, value.iri))
        for value in values.get(Literal, set()):
            self.session.graph.add((self.iri, annotation.iri, value))
        for value in values.get(ATTRIBUTE_VALUE_TYPES, set()):
            self.session.graph.add((self.iri, annotation.iri, Literal(value)))
        for value in values.get(URIRef, set()):
            self.session.graph.add((self.iri, annotation.iri, value))

    def annotations_delete(
        self,
        annotation: OntologyAnnotation,
        values: Union[
            Dict[Type[AnnotationValue], Union[AnnotationValue]],
            Set[AnnotationValue],
        ],
    ) -> None:
        """Deletes an annotation from an individual."""
        if not isinstance(values, dict):
            values = self._classify_by_type(values)

        for value in values.get(Literal, set()):
            self.session.graph.remove((self.iri, annotation.iri, value))
        for value in values.get(ATTRIBUTE_VALUE_TYPES, set()):
            self.session.graph.remove(
                (self.iri, annotation.iri, Literal(value))
            )
        for value in values.get(URIRef, set()):
            self.session.graph.remove((self.iri, annotation.iri, value))

    def annotations_set(
        self,
        annotation: OntologyAnnotation,
        values: Union[
            Dict[Type[AnnotationValue], AnnotationValue], Set[AnnotationValue]
        ],
    ) -> None:
        """Replaces the annotations of an individual."""
        if not isinstance(values, dict):
            values = self._classify_by_type(values)

        self.session.graph.remove((self.iri, annotation.iri, None))
        for value in chain(
            *(
                values.get(key, set())
                for key in (
                    OntologyAnnotation,
                    OntologyAttribute,
                    OntologyClass,
                    OntologyIndividual,
                    OntologyRelationship,
                )
            )
        ):
            self.session.graph.add((self.iri, annotation.iri, value.iri))
        for value in values.get(Literal, set()):
            self.session.graph.add((self.iri, annotation.iri, value))
        for value in values.get(ATTRIBUTE_VALUE_TYPES, set()):
            self.session.graph.add((self.iri, annotation.iri, Literal(value)))
        for value in values.get(URIRef, set()):
            self.session.graph.add((self.iri, annotation.iri, value))

    def annotations_value_generator(
        self, annotation: OntologyAnnotation
    ) -> Iterator[AnnotationValue]:
        """Yields the annotation values applied to the individual."""
        for obj in self.session.graph.objects(self.iri, annotation.iri):
            if isinstance(obj, URIRef):
                try:
                    yield self.session.from_identifier(obj)
                    continue
                except KeyError:
                    pass
                try:
                    yield self.session.ontology.from_identifier(obj)
                    continue
                except KeyError:
                    pass
                yield obj
            elif isinstance(obj, Literal):
                yield obj.toPython()

    def annotations_iter(
        self,
        rel: Optional[OntologyAnnotation] = None,
        return_rel: bool = False,
    ) -> Iterator[AnnotationValue]:
        """Iterate over the connected ontology individuals.

        Args:
            rel: Only return the annotation values which are connected by
                the given annotation. Defaults to None (any relationship).
            return_rel: Whether to return the connecting
                relationship. Defaults to False.

        Returns:
            Iterator with the queried ontology individuals.
        """
        entities_and_annotations = (
            (
                self.session.from_identifier(o),
                self.session.ontology.from_identifier(p),
            )
            for s, p, o in self.session.graph.triples(
                (
                    self.identifier,
                    rel.identifier if rel is not None else None,
                    None,
                )
            )
            if not (isinstance(o, Literal) or p == RDF.type)
        )
        entities_and_annotations = filter(
            lambda x: isinstance(x, OntologyAnnotation),
            entities_and_annotations,
        )
        if return_rel:
            yield from entities_and_annotations
        else:
            yield from map(lambda x: x[0], entities_and_annotations)

    # ↑ --------------- ↑
    # Annotation handling

    # Attribute handling
    # ↓ -------------- ↓

    def _attributes_get_by_name(self, name: str) -> Set[OntologyAttribute]:
        """Get an attribute of this individual by name."""
        attributes = (
            attr
            for oclass in self.classes
            for attr in chain(
                oclass.attributes.keys(),
                oclass.optional_attributes,
                self.attributes_generator(),
            )
        )
        attributes = (
            attr
            for attr in attributes
            if any(
                (
                    str(attr.identifier).endswith(name),
                    name in attr.iter_labels(return_literal=False),
                )
            )
        )
        return set(attributes)

    def attributes_add(
        self, attribute: OntologyAttribute, values: Iterable[AttributeValue]
    ):
        """Add values to a datatype property.

        If any of the values provided in `values` have already been assigned,
        then they are simply ignored.

        Args:
            attribute: The ontology attribute to be used for assignments.
            values: An iterable of objects whose types are compatible either
                with the OWL standard's data types for literals or compatible
                with SimPhoNy as custom data types.

        Raises:
            TypeError: When Python objects with types incompatible with the
                OWL standard or with SimPhoNy as custom data types are given.
        """
        # TODO: prevent the end result having more than one value depending on
        #  ontology cardinality restrictions and/or functional property
        #  criteria.
        values = (
            attribute.convert_to_datatype(value)
            for value in values
            if value is not None
        )
        for value in values:
            self.session.graph.add(
                (
                    self.iri,
                    attribute.iri,
                    Literal(
                        value,
                        datatype=attribute.datatype,
                    ),
                )
            )

    def attributes_delete(
        self, attribute: OntologyAttribute, values: Iterable[AttributeValue]
    ):
        """Remove values from a datatype property.

        If any of the values provided in `values` are not present, they are
        simply ignored.

        Args:
            attribute: The ontology attribute to be used for assignments.
            values: An iterable of objects whose types are compatible either
                with the OWL standard's data types for literals or compatible
                with SimPhoNy as custom data types.

        Raises:
            TypeError: When Python objects with types incompatible with the
                OWL standard or with SimPhoNy as custom data types are given.
        """
        values = (
            attribute.convert_to_datatype(value)
            for value in values
            if value is not None
        )
        for value in values:
            self.session.graph.remove(
                (
                    self.iri,
                    attribute.iri,
                    Literal(
                        value,
                        datatype=attribute.datatype,
                    ),
                )
            )

    def attributes_set(
        self,
        attribute: OntologyAttribute,
        values: Iterable[Union[AttributeValue, Literal]],
    ):
        """Replace values assigned to a datatype property.

        Args:
            attribute: The ontology attribute to be used for assignments.
            values: An iterable of objects whose types are compatible either
                with the OWL standard's data types for literals or compatible
                with SimPhoNy as custom data types.

        Raises:
            TypeError: When Python objects with types incompatible with the
                OWL standard or with SimPhoNy as custom data types are given.
        """
        # TODO: prevent the end result having more than one value depending on
        #  ontology cardinality restrictions and/or functional property
        #  criteria.
        self.session.graph.remove((self.iri, attribute.iri, None))
        self.attributes_add(attribute, values)

    def attributes_value_generator(
        self, attribute: OntologyAttribute
    ) -> Iterator[AttributeValue]:
        """Returns a generator of values assigned to the specified attribute.

        Args:
            attribute: The ontology attribute query for values.

        Returns:
            Generator that returns the attribute values.
        """
        for literal in self.session.graph.objects(self.iri, attribute.iri):
            # TODO: Recreating the literal to get a vector from
            #  literal.toPython() should not be necessary, find out why it
            #  is happening.
            literal = Literal(
                str(literal), datatype=literal.datatype, lang=literal.language
            )
            yield literal.toPython()

    def attributes_value_contains(
        self, attribute: OntologyAttribute, value: AttributeValue
    ) -> bool:
        """Whether a specific value is assigned to the specified attribute.

        Args:
            attribute: The ontology attribute query for values.

        Returns:
            Whether the specific value is assigned to the specified
            attribute or not.
        """
        if attribute.datatype in (None, RDF.langString):
            return any(
                str(value) == str(x)
                for x in self.session.graph.objects(self.iri, attribute.iri)
                if isinstance(x, Literal)
            )
        else:
            literal = Literal(value, datatype=attribute.datatype)
            literal = Literal(str(literal), datatype=attribute.datatype)
            return literal in self.session.graph.objects(
                self.iri, attribute.iri
            )

    def attributes_generator(self) -> Iterator[OntologyAttribute]:
        """Returns a generator of the attributes of this CUDS object.

        The generator only returns the OntologyAttribute objects, NOT the
        values.

        Returns:
            Generator that returns the attributes of this CUDS object.
        """
        for predicate in self.session.graph.predicates(self.iri, None):
            try:
                obj = self.session.ontology.from_identifier_typed(
                    predicate, typing=OntologyAttribute
                )
            except (KeyError, TypeError):
                continue
            if isinstance(obj, OntologyAttribute):
                yield obj

    def attributes_attribute_and_value_generator(
        self,
    ) -> Iterator[Tuple[OntologyAttribute, Iterator[AttributeValue]]]:
        """Returns a generator of both the attributes and their values.

        Returns:
            Generator that yields tuples, where the first item is the ontology
            attribute and the second a generator of values for such attribute.
        """
        for attribute in self.attributes_generator():
            yield attribute, self.attributes_value_generator(attribute)

    # ↑ -------------- ↑
    # Attribute handling

    # Relationship handling
    # ↓ ----------------- ↓

    def relationships_iter(
        self,
        rel: Optional[OntologyRelationship] = None,
        oclass: Optional[OntologyClass] = None,
        return_rel: bool = False,
    ) -> Union[
        Iterator[OntologyIndividual],
        Iterator[Tuple[OntologyIndividual, OntologyRelationship]],
    ]:
        """Iterate over the connected ontology individuals.

        Args:
            rel: Only return ontology individuals which are connected by
                the given relationship. Defaults to None (any relationship).
            oclass: Only return ontology individuals which belong to the
                given ontology class. Defaults to None (any class).
            return_rel: Whether to return the connecting
                relationship. Defaults to False.

        Returns:
            Iterator with the queried ontology individuals.
        """

        def individuals_and_relationships() -> Iterator[
            OntologyIndividual, OntologyEntity
        ]:
            for s, p, o in self.session.graph.triples(
                (
                    self.identifier,
                    rel.identifier if rel is not None else None,
                    None,
                )
            ):
                if isinstance(o, Literal) or p == RDF.type:
                    continue
                prop = self.session.ontology.from_identifier(p)
                if not isinstance(prop, OntologyRelationship):
                    continue
                individual = self.session.from_identifier_typed(
                    o, typing=OntologyIndividual
                )
                yield individual, prop

        individuals_and_relationships = individuals_and_relationships()
        if oclass:
            individuals_and_relationships = (
                (entity, relationship)
                for entity, relationship in individuals_and_relationships
                if oclass == entity
            )

        if return_rel:
            yield from individuals_and_relationships
        else:
            yield from map(lambda x: x[0], individuals_and_relationships)

    # ↑ ----------------- ↑
    # Relationship handling



"""Utility functions and variables for the remote store implementation."""

import hashlib
import urllib.parse
from enum import Enum
from typing import Optional, Tuple


class COMMAND(str, Enum):
    """Collection of remote interface commands."""

    # Interface commands
    ROOT = "ROOT"
    OPEN = "OPEN"
    CLOSE = "CLOSE"
    POPULATE = "POPULATE"
    COMMIT = "COMMIT"
    COMPUTE = "COMPUTE"

    # Base graph commands
    STORE_OPEN = "STORE_OPEN"
    STORE_CLOSE = "STORE_CLOSE"
    STORE_ADD = "STORE_ADD"
    STORE_TRIPLES = "STORE_TRIPLES"
    STORE_REMOVE = "STORE_REMOVE"
    STORE_COMMIT = "STORE_COMMIT"
    STORE_ROLLBACK = "STORE_ROLLBACK"

    # Triplestore commands
    ADD = "ADD"
    TRIPLES = "TRIPLES"
    REMOVE = "REMOVE"

    # File commands
    SAVE = "SAVE"
    LOAD = "LOAD"
    DELETE = "DELETE"
    HASH = "HASH"
    RENAME = "RENAME"

    # Remote interface commands
    HASATTR = "HASATTR"
    AUTHENTICATE = "AUTHENTICATE"


def get_hash(file_path: str) -> str:
    """Get the hash of the given file.

    Args:
        file_path (path): A path to a file

    Returns:
        HASH: A sha256 HASH object
    """
    buf_size = 4096
    result = hashlib.sha256()
    with open(file_path, "rb") as f:
        data = True
        while data:
            data = f.read(buf_size)
            result.update(data)
    return result.hexdigest()


def parse_uri(uri: str) -> Tuple[Optional[str], Optional[str], Optional[str]]:
    """Parse the given uri and return uri, username, password.

    Args:
        uri: The URI to parse
    """
    if uri is None:
        return None, None, None
    parsed = urllib.parse.urlparse(uri)
    username = parsed.username
    password = parsed.password
    parsed = list(parsed)
    if username or password:
        parsed[1] = parsed[1].split("@")[1]
    uri = urllib.parse.urlunparse(parsed)
    return uri, username, password



"""Tools aimed at remote interfaces."""

from typing import Iterable, Optional, Type, Union

from simphony_osp.interfaces.remote.server import InterfaceServer
from simphony_osp.session.wrapper import WrapperSpawner


def host(
    wrapper: Type[WrapperSpawner],
    configuration_string: str = "",
    create: bool = False,
    hostname: str = "127.0.0.1",
    port: int = "6537",
    username: Optional[str] = None,
    password: Optional[str] = None,
    **kwargs: Union[
        str,
        int,
        float,
        bool,
        None,
        Iterable[Union[str, int, float, bool, None]],
    ]
):
    """Host a server based on a wrapper.

    Opens the specified wrapper and starts listening for clients. The
    clients can connect to the wrapper's session and perform actions.

    Args:
        wrapper: The wrapper to be used.
        configuration_string: The configuration string of the wrapper.
        create: The value of the argument `create` for the wrapper.
        hostname: Hostname where the server will listen.
        port: The port that the server will use to listen.
        username: A username for authenticating the client.
        password: A password for authenticating the client.
        **kwargs: Keyword arguments for the wrapper.
    """

    def _interface_generator(
        user: Optional[str] = None, pass_: Optional[str] = None
    ):
        if username and user != username:
            raise PermissionError
        if password and pass_ != pass_:
            raise PermissionError
        session = wrapper(configuration_string, create, **kwargs)
        interface = session.driver.interface
        return interface

    interface_server = InterfaceServer(
        host=hostname, port=port, generate_interface=_interface_generator
    )

    interface_server.listen()



"""Module containing tools for the users of the SimPhoNy OSP."""

import simphony_osp.utils.pico as pico
from simphony_osp.tools.general import branch, relationships_between
from simphony_osp.tools.import_export import export_file, import_file
from simphony_osp.tools.pretty_print import pretty_print
from simphony_osp.tools.remote import host
from simphony_osp.tools.search import (
    find,
    find_by_attribute,
    find_by_class,
    find_by_identifier,
    find_relationships,
    sparql,
)
from simphony_osp.tools.semantic2dot import semantic2dot

__all__ = [
    # simphony_osp.tools.import_export
    "export_file",
    "import_file",
    # simphony_osp.tools.pico
    "pico",
    # simphony_osp.tools.pretty_print
    "pretty_print",
    # simphony_osp.tools.search
    "sparql",
    "find",
    "find_by_attribute",
    "find_by_class",
    "find_by_identifier",
    "find_relationships",
    # simphony_osp.tools.semantic2dot
    "semantic2dot",
    # simphony_osp.tools.general
    "branch",
    "relationships_between",
    # simphony_osp.utils.remote
    "host",
]



"""Wrappers are interfaces between the SimPhoNy OSP and other software.

This module lets the user access the SimPhoNy wrappers that are installed.

On the `setup.py` file of each wrapper, the wrapper class is registered with
the [entry point](https://setuptools.pypa.io/en/latest/userguide/entry_point
.html#advertising-behavior) `simphony_osp.wrappers`. This module loads all
the classes registered with this entry point and makes them accessible
within its namespace. It also makes them discoverable through IPython tab
completion.
"""

import sys
from typing import Type

from simphony_osp.interfaces.interface import Interface
from simphony_osp.session.wrapper import WrapperSpawner as Wrapper

if sys.version_info < (3, 8):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

# Retrieve all wrappers from package entry points.
package_entry_points = entry_points()
if sys.version_info >= (3, 10):
    wrappers = package_entry_points.select(group="simphony_osp.wrappers")
else:
    wrappers = package_entry_points.get("simphony_osp.wrappers", tuple())
del package_entry_points
wrappers = {entry_point.name: entry_point.load() for entry_point in wrappers}


def __getattr__(name: str):
    """Retrieve a wrapper.

    The wrapper is an object that the user can call to spawn a
    wrapper ontology entity.

    Args:
        name: Name of the wrapper. The name of a wrapper is the name chosen
            by the developer for the entry point.
    """
    if name not in wrappers:
        raise AttributeError(name)

    class WrapperSpawner(Wrapper):
        """Adds a method to retrieve the Interface class of a wrapper.

        Internally, SimPhoNy calls the class that the wrapper developer
        implements "Interface" class. The "Wrapper" class from which this
        one inherits just provides a way for the user to start using a wrapper.

        This specific "WrapperSpawner" class just defines the abstract method
        "_get_interface" that the "Wrapper" class requires.
        """

        @classmethod
        def _get_interface(cls) -> Type[Interface]:
            return wrappers[name]

    return WrapperSpawner


def __dir__():
    return list(wrappers.keys())



"""The data space store connects SimPhoNy to a data space."""
import os
import pathlib
from base64 import b64encode
from pathlib import Path
from typing import BinaryIO, Optional

from rdflib import Graph, URIRef
from rdflib.term import Identifier

from simphony_osp.interfaces.interface import Interface
from simphony_osp.interfaces.remote.common import get_hash


class DataspaceInterface(Interface):
    """The data space interface connects SimPhoNy to a data space."""

    _identifier: Identifier = URIRef(
        "https://www.simphony-project.eu/SQLAlchemy"
    )

    _database_path: Optional[Path] = None
    _files_path: Optional[Path] = None

    _uri: Optional[str] = None

    # Interface
    # ↓ ----- ↓

    entity_tracking: bool = False

    def open(self, configuration: str, create: bool = False):
        """Open the specified dataspace."""
        path = pathlib.Path(configuration).absolute()
        if not create:
            if not path.is_dir():
                raise FileNotFoundError(f"Folder {path} not found")
            if not (path / "database.db").is_file():
                raise FileNotFoundError(
                    f'Database {path / "database.db"} ' f"not found."
                )
            if not (path / "files").is_dir():
                raise FileNotFoundError(
                    f'Folder {path / "files"} ' f"not found."
                )

        uri = "sqlite:///" + str(path / "database.db")
        if self._uri is not None and self._uri != uri:
            raise RuntimeError(
                f"A different dataspace {self._uri}" f"is already open!"
            )

        os.makedirs(path, exist_ok=True)
        os.makedirs(path / "files", exist_ok=True)
        self.base = Graph("SQLAlchemy", identifier=self._identifier)
        self.base.open(uri, create=create)
        self._uri = uri
        self._database_path = path / "database.db"
        self._files_path = path / "files"

    def close(self):
        """Close the dataspace."""
        if self.base is not None:
            self.base.close(commit_pending_transaction=False)
            self._uri = None
            self.base = None
            self._database_path = None
            self._files_path = None

    def commit(self):
        """Commit pending changes to the triple store."""
        # The `InterfaceDriver` will simply add the triples to the base graph
        # and commit them. Nothing to do here.
        pass

    def populate(self):
        """The base graph does not need to be populated. Nothing to do."""
        pass

    def save(self, key: str, file: BinaryIO) -> None:
        """Save a file."""
        file_name = b64encode(bytes(key, encoding="UTF-8")).decode("UTF-8")
        buf_size = 1024
        with open(self._files_path / file_name, "wb") as new_file:
            data = True
            while data:
                data = file.read(buf_size)
                new_file.write(data)

    def load(self, key: str) -> BinaryIO:
        """Load a file."""
        file_name = b64encode(bytes(key, encoding="UTF-8")).decode("UTF-8")
        return open(self._files_path / file_name, "rb")

    def delete(self, key: str) -> None:
        """Delete a file."""
        file_name = b64encode(bytes(key, encoding="UTF-8")).decode("UTF-8")
        (self._files_path / file_name).unlink()

    def hash(self, key: str) -> str:
        """Hash a file."""
        file_name = b64encode(bytes(key, encoding="UTF-8")).decode("UTF-8")
        return get_hash(str(self._files_path / file_name))

    def rename(self, key: str, new_key: str) -> None:
        """Rename a file."""
        file_name = b64encode(bytes(key, encoding="UTF-8")).decode("UTF-8")
        new_file_name = b64encode(bytes(new_key, encoding="UTF-8")).decode(
            "UTF-8"
        )
        (self._files_path / file_name).rename(self._files_path / new_file_name)

    # ↑ ----- ↑



"""An attribute defined in the ontology."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Iterable, Iterator, Optional

from rdflib import OWL, RDFS, Literal, URIRef
from rdflib.term import Identifier

from simphony_osp.ontology.entity import OntologyEntity
from simphony_osp.utils.datatypes import UID, Triple, rdf_to_python

if TYPE_CHECKING:
    from simphony_osp.session.session import Session

logger = logging.getLogger(__name__)


class OntologyAttribute(OntologyEntity):
    """An attribute defined in the ontology."""

    rdf_type = OWL.DatatypeProperty
    rdf_identifier = Identifier

    # Public API
    # ↓ ------ ↓

    @property
    def datatype(self) -> Optional[URIRef]:
        """Get the data type of the attribute.

        Returns:
            IRI of the datatype.

        Raises:
            NotImplementedError: More than one data type associated with the
                attribute.
        """
        data_types = (
            o
            for superclass in self.superclasses
            for o in self.session.graph.objects(superclass.iri, RDFS.range)
        )
        result = next(data_types, None)
        if next(data_types, None) is not None:
            raise NotImplementedError(
                f"More than one datatype associated to {self}: {data_types}."
            )
        return result

    # ↑ ------ ↑
    # Public API

    def __init__(
        self,
        uid: UID,
        session: Optional[Session] = None,
        triples: Optional[Iterable[Triple]] = None,
        merge: bool = False,
    ) -> None:
        """Initialize the ontology attribute.

        Args:
            uid: UID identifying the entity.
            session: Session where the entity is stored.
            triples: Construct the attribute with the provided triples.
            merge: Whether overwrite the potentially existing entity in the
                session with the provided triples or just merge them with
                the existing ones.
        """
        super().__init__(uid, session, triples, merge=merge)

    def convert_to_datatype(self, value: Any) -> Any:
        """Convert the given value to a Python object.

        The class of the Python object depends on the data type of the
        attribute.

        Args:
            value: The value to convert.

        Returns:
            The converted value.
        """
        python_object = rdf_to_python(value, self.datatype)
        if isinstance(python_object, Literal):
            raise TypeError(
                f"Type '{type(value)}' of object {value} cannot be set as "
                f"attribute value, as it is either incompatible with the "
                f"OWL standard or not yet supported by SimPhoNy."
            )
        return python_object

    def _get_direct_superclasses(self) -> Iterator[OntologyAttribute]:
        """Get all the direct superclasses of this attribute.

        Returns:
            The direct superattributes.
        """
        for o in self.session.graph.objects(self.iri, RDFS.subPropertyOf):
            try:
                yield self.session.from_identifier_typed(
                    o, typing=OntologyAttribute
                )
            except TypeError:
                pass
        # The try-catch block makes sure that `OntologyAnnotation` and
        #  `OntologyRelationship` objects are not superclasses, as
        #  `RDFS.subPropertyOf` is used to establish class hierarchies of
        #  rdf:Property, owl:DatatypeProperty, owl:ObjectProperty and
        #  owl:AnnotationProperty.

    def _get_direct_subclasses(self) -> Iterator[OntologyAttribute]:
        """Get all the direct subclasses of this attribute.

        Returns:
            The direct subattributes.
        """
        for o in self.session.graph.subjects(RDFS.subPropertyOf, self.iri):
            try:
                yield self.session.from_identifier_typed(
                    o, typing=OntologyAttribute
                )
            except TypeError:
                pass
        # The try-catch makes sure that `OntologyAnnotation` and
        #  `OntologyRelationship` objects are not superclasses, as
        #  `RDFS.subPropertyOf` is used to establish class hierarchies of
        #  rdf:Property, owl:DatatypeProperty, owl:ObjectProperty and
        #  owl:AnnotationProperty.

    def _get_superclasses(self) -> Iterator[OntologyAttribute]:
        """Get all the superclasses of this attribute.

        Yields:
            The super-attributes.
        """
        yield self

        def closure(node, graph):
            yield from graph.objects(node, RDFS.subPropertyOf)

        for x in self.session.graph.transitiveClosure(
            closure, self.identifier
        ):
            try:
                yield self.session.from_identifier_typed(
                    x, typing=OntologyAttribute
                )
            except TypeError:
                pass
        # The filter makes sure that `OntologyAnnotation` and
        #  `OntologyRelationship` objects are not superclasses, as
        #  `RDFS.subPropertyOf` is used to establish class hierarchies of
        #  rdf:Property, owl:DatatypeProperty, owl:ObjectProperty and
        #  owl:AnnotationProperty.

        yield self.session.from_identifier(OWL.topDataProperty)

    def _get_subclasses(self) -> Iterator[OntologyAttribute]:
        """Get all the subclasses of this attribute.

        Yields:
            The sub-attributes.
        """
        yield self

        def closure(node, graph):
            yield from graph.subjects(RDFS.subPropertyOf, node)

        for x in self.session.graph.transitiveClosure(
            closure, self.identifier
        ):
            try:
                yield self.session.from_identifier_typed(
                    x, typing=OntologyAttribute
                )
            except TypeError:
                pass
        # The filter makes sure that `OntologyAnnotation` and
        #  `OntologyRelationship` objects are not superclasses, as
        #  `RDFS.subPropertyOf` is used to establish class hierarchies of
        #  rdf:Property, owl:DatatypeProperty, owl:ObjectProperty and
        #  owl:AnnotationProperty.
